"""
scanpy dpt
"""

import numpy as np
import scanpy as sc
from ..obj_utils import (
    _rename_default_key,
)


def dpt(
        adata,
        root=None,
        use_diffmap='X_diffmap',
        key_added=None,
        **kwargs,
):
    """
    Wrapper function for sc.tl.dpt
    """
    if root is None or not (isinstance(root, (list, tuple)) and len(root) == 2):
        root = (None, None)
    if 'iroot' not in adata.uns.keys() and root[0] is None:
        raise ValueError('Annotate your data with root cell first, i.e. '
                         'boolean vector `.uns["iroot"]` is required.')
    if root[0] is not None:
        adata.uns['iroot'] = np.random.choice(
            np.flatnonzero(adata.obs[root[0]] == root[1]))

    sc.tl.dpt(adata, **kwargs)
    if key_added:
        dpt_key = f'dpt_pseudotime_{key_added}'
        _rename_default_key(adata.obs, 'dpt_pseudotime', dpt_key)
    
    return adata



"""
scanpy external bbknn
"""

import scanpy.external as sce

from ..obj_utils import (
    _backup_default_key,
    _delete_backup_key,
    _rename_default_key,
)

# Wrapper for bbknn allowing use of non-standard slot

def bbknn(adata, key=None, key_added=None, **kwargs):
    """
    Wrapper function for sce.pp.bbknn(), for supporting non-standard neighbors slot
    """

    _backup_default_key(adata.uns, 'neighbors')
    _backup_default_key(adata.obsp, 'distances')
    _backup_default_key(adata.obsp, 'connectivities')
    sce.pp.bbknn(adata, batch_key = key, **kwargs)    

    if key_added:
        _rename_default_key(adata.uns, 'neighbors',  f'{key_added}')
        _rename_default_key(adata.obsp, 'distances',  f'{key_added}_distances')
        _rename_default_key(adata.obsp, 'connectivities',  f'{key_added}_connectivities')
    else:
        _delete_backup_key(adata.uns, 'neighbors')
        _delete_backup_key(adata.obsp, 'distances')
        _delete_backup_key(adata.obsp, 'connectivities')
    
    return adata



"""
Provide cmd options
"""

import click
from .click_utils import (
    CommaSeparatedText,
    Dictionary,
    valid_limit,
    valid_parameter_limits,
    mutually_exclusive_with,
    required_by,
)

COMMON_OPTIONS = {
    'input': [
        click.argument(
            'input_obj',
            metavar='<input_obj>',
            type=click.Path(exists=True, dir_okay=False),
        ),
        click.option(
            '--input-format', '-f',
            type=click.Choice(['anndata', 'loom']),
            default='anndata',
            show_default=True,
            help='Input object format.',
        ),
    ],

    'output': [
        click.argument(
            'output_obj',
            metavar='<output_obj>',
            type=click.Path(dir_okay=False, writable=True),
        ),
        click.option(
            '--output-format', '-F',
            type=click.Choice(['anndata', 'loom', 'zarr']),
            default='anndata',
            show_default=True,
            help='Output object format.',
        ),
        click.option(
            '--zarr-chunk-size', '-z',
            type=click.INT,
            default=1000,
            show_default=True,
            help='Chunk size for writing output in zarr format.',
        ),
        click.option(
            '--loom-write-obsm-varm', '-b',
            is_flag=True,
            default=False,
            show_default=True,
            help='Write obsm and varm to the Loom file?',
        ),
        click.option(
            '--export-mtx', '-X',
            type=click.Path(dir_okay=True, writable=True),
            default=None,
            show_default=True,
            help='When specified, using it as prefix for exporting mtx files. '
            'If not empty and not ending with "/" or "_", a "_" will be '
            'appended.',
        ),
        click.option(
            '--show-obj',
            type=click.Choice(['stdout', 'stderr']),
            default=None,
            show_default=True,
            help='Print output object summary info to specified stream.',
        ),
    ],

    'save': [
        click.option(
            '--save-raw', '-r',
            is_flag=True,
            default=False,
            show_default=True,
            help='Save adata to adata.raw before processing.',
        ),
        click.option(
            '--save-layer', '-y',
            type=click.STRING,
            default=None,
            show_default=True,
            help='Save adata.X to the specified layer before processing.',
        ),
    ],

    'plot': [
        click.argument(
            'output_fig',
            metavar='<output_fig>',
            type=click.Path(dir_okay=False, writable=True),
        ),
        click.option(
            '--fig-size',
            type=CommaSeparatedText(click.INT, length=2),
            default="7,7",
            show_default=True,
            help='Figure size.',
        ),
        click.option(
            '--fig-dpi',
            type=click.INT,
            default=80,
            show_default=True,
            help='Figure DPI.',
        ),
        click.option(
            '--fig-fontsize',
            type=click.INT,
            default=15,
            show_default=True,
            help='Figure font size.',
        ),
    ],

    'frame_title': [
        click.option(
            '--frameon/--frameoff', 'frameon',
            default=True,
            show_default=True,
            help='Draw a frame around the plot',
        ),
        click.option(
            '--title',
            type=CommaSeparatedText(simplify=True),
            default=None,
            show_default=True,
            help='Provide title for the plot or panels.',
        ),
    ],

    'use_pc': [
        click.option(
            '--n-pcs', '-n',
            type=click.INT,
            default=None,
            show_default=True,
            help='Use this many PCs. Use `.X` if --n-pcs is 0 when --use-rep is '
            'None.',
        ),

        click.option(
            '--use-rep', '-u',
            type=click.STRING,
            default=None,
            show_default=True,
            help='Use the indicated representation. If None, the representation is '
            'chosen automatically: for `.n_vars` < 50, `.X` is used, otherwise '
            '`X_pca` is used. If `X_pca` is not present, it\'s computed with '
            'default parameters.'
        ),
    ],

    'knn_graph': [
        click.option(
            '--neighbors-key',
            type=click.STRING,
            default=None,
            show_default=False,
            help='If not specified, look in .uns[‘neighbors’] for neighbors '
            'settings and .obsp[‘connectivities’], .obsp[‘distances’] for connectivities and '
            'distances respectively (default storage places for pp.neighbors). If specified, '
            'look in .uns[neighbors_key] for neighbors settings and '
            '.obsp[.uns[neighbors_key][‘connectivities_key’]], '
            '.obsp[.uns[neighbors_key][‘distances_key’]] for connectivities and distances '
            'respectively.'
        ),
        click.option(
            '--obsp',
            type=click.STRING,
            default=None,
            show_default=True,
            help='Use .obsp[obsp] as adjacency. You can’t specify both obsp and '
            'neighbors_key at the same time.'
        ),
        click.option(
            '--directed/--undirected', 'directed',
            default=True,
            show_default=True,
            help='Interpret the adjacency matrix as directed graph.',
        ),
        click.option(
            '--use-weights',
            is_flag=True,
            default=False,
            show_default=True,
            help='Use weights from KNN graph.',
        ),
    ],

    
    'layer':click.option(
        '--layer',
        type=CommaSeparatedText(simplify=True),
        default=None,
        show_default=True,
        help='Name of the AnnData object layer that wants to be plotted. By '
        'default adata.raw.X is plotted. If use_raw=False is set, then adata.X '
        'is plotted. If layer is set to a valid layer name, then the layer is '
        'plotted. layer takes precedence over use_raw.',
    ),

    'n_comps': click.option(
        '--n-comps',
        type=click.INT,
        default=None,
        show_default=True,
        help='Number of components to compute',
    ),

    'key_added': click.option(
        '--key-added',
        type=CommaSeparatedText(simplify=True),
        default=None,
        show_default=True,
        help='Key under which to add the computed results',
    ),

    'random_state': click.option(
        '--random-state', '-S',
        type=click.INT,
        default=0,
        show_default=True,
        help='Seed for random number generator.',
    ),

    'use_raw': click.option(
        '--use-raw/--no-raw', 'use_raw',
        default=True,
        show_default=True,
        help='Use expression values in `.raw` if present.',
    ),

    'zero_center': click.option(
        '--no-zero-center', 'zero_center',
        is_flag=True,
        flag_value=False,
        default=True,
        help='When set, omit zero-centering variables to allow efficient '
        'handling of sparse input.',
    ),

    'n_jobs': click.option(
        '--n-jobs', '-J',
        type=click.INT,
        default=None,
        show_default=True,
        help='Number of jobs for parallel computation.',
    ),

    'restrict_to': click.option(
        '--restrict-to',
        type=(click.STRING, CommaSeparatedText()),
        default=(None, None),
        show_default=True,
        help='Restrict the clustering to the categories within the key for '
        'sample annotation, in the form of "obs_key list_of_categories".',
    ),

    'export_embedding': click.option(
        '--export-embedding', '-E',
        type=click.Path(dir_okay=False, writable=True),
        default=None,
        show_default=True,
        help='Export embeddings in a tab-separated text table.',
    ),

    'export_cluster': click.option(
        '--export-cluster',
        type=click.Path(dir_okay=False, writable=True),
        default=None,
        show_default=True,
        help='Export embeddings in a tab-separated text table.',
    ),

    'var_names': click.option(
            '--var-names',
            type=(CommaSeparatedText()),
            show_default=True,
            help='var_names should be a valid subset of adata.var_names.',
    ),

    'gene_symbols': click.option(
            '--gene-symbols',
            type=CommaSeparatedText(simplify=True),
            default=None,
            show_default=True,
            help='Column name in .var DataFrame that stores gene symbols. By '
            'default this is assumed to be the index column of the .var '
            'DataFrame. Setting this option allows alternative names to be '
            'used.',
    ),

    'diffexp_plot': [
        click.option(
            '--rgg',
            is_flag=True,
            default=False,
            show_default=True,
            help='When set, use the rank_genes_groups_ form of the function, '
            'where gene lists are automatically selected.',
        ),
        click.option(
             '--groupby',
            type=CommaSeparatedText(simplify=True),
            default=None,
            show_default=True,
            help='The key of the observation grouping to consider.',
        ),
        click.option(
            '--log',
            is_flag=True,
            default=False,
            show_default=True,
            help='Plot on logarithmic axis.',
        ),
        click.option(
            '--num-categories',
            type=click.INT,
            default=7,
            show_default=True,
            help='Only used if groupby observation is not categorical. This value '
            'determines the number of groups into which the groupby observation '
            'should be subdivided.',
        ),
        click.option(
            '--dendrogram',
            is_flag=True,
            default=False,
            show_default=False,
            help='If True, a dendrogram based on the hierarchical clustering '
            'between the groupby categories is added. The dendrogram information is '
            'computed using scanpy.tl.dendrogram(). If tl.dendrogram has not been '
            'called previously the function is called with default parameters.',
        ),
        click.option(
            '--standard-scale',
            type=click.Choice(['var', 'obs']),
            default=None,
            show_default=True,
            help='Whether or not to standardize that dimension between 0 and 1, '
            'meaning for each variable or group, subtract the minimum and divide '
            'each by its maximum.'
        ),

    ],

    'sviol': [
        click.option(
            '--no-stripplot', 'stripplot',
            is_flag=True,
            default=True,
            show_default=True,
            help='When set, do not add a stripplot on top of the violin plot.',
        ),
        click.option(
            '--no-jitter', 'jitter',
            is_flag=True,
            default=True,
            show_default=True,
            help='Suppress jitter in the stripplot (only when stripplot is True)'
        ),
        click.option(
            '--size',
            type=click.INT,
            default=1,
            show_default=True,
            help='Size of the jitter points.'
        ),
        click.option(
            '--order',
            type=CommaSeparatedText(),
            default=None,
            show_default=True,
            help='Order in which to show the categories.'
        ),
        click.option(
            '--scale',
            type=click.Choice(['area', 'count', 'width']),
            default='width',
            show_default=True,
            help='The method used to scale the width of each violin. If ‘area’, '
            'each violin will have the same area. If ‘count’, the width of the '
            'violins will be scaled by the number of observations in that bin. If '
            '‘width’, each violin will have the same width.'
        ),
        click.option(
            '--row-palette',
            type=CommaSeparatedText(simplify=True),
            default='muted',
            show_default=True,
            help='The row palette determines the colors to use in each of the '
            'stacked violin plots. The value should be a valid seaborn palette name '
            'or a valic matplotlib colormap (see '
            'https://seaborn.pydata.org/generated/seaborn.color_palette.html). '
            'Alternatively, a single color name or hex value can be passed. E.g. '
            '‘red’ or ‘#cc33ff’.'
        ),
    ],


    'dot': [
        click.option(
            '--expression-cutoff',
            type=click.FLOAT,
            default=0,
            show_default=True,
            help='Expression cutoff that is used for binarizing the gene expression '
            'and determining the fraction of cells expressing given genes. A gene is '
            'expressed only if the expression value is greater than this threshold.'
        ),
        click.option(
            '--mean-only-expressed',
            is_flag=True,
            default=False,
            show_default=True,
            help='If True, gene expression is averaged only over the cells '
            'expressing the given genes.',
        ),
        click.option(
            '--color-map',
            type=CommaSeparatedText(simplify=True),
            default='Reds',
            show_default=True,
            help='String denoting matplotlib color map.',
        ),
        click.option(
            '--dot-max',
            type=click.FLOAT,
            default=None,
            show_default=True,
            help='If none, the maximum dot size is set to the maximum fraction '
            'value found (e.g. 0.6). If given, the value should be a number between '
            '0 and 1. All fractions larger than dot_max are clipped to this value.'
        ),
        click.option(
            '--dot-min',
            type=click.FLOAT,
            default=None,
            show_default=True,
            help='If none, the minimum dot size is set to 0. If given, the value '
            'should be a number between 0 and 1. All fractions smaller than dot_min '
            'are clipped to this value.'
        ),
        click.option(
            '--smallest-dot',
            type=click.FLOAT,
            default=0,
            show_default=True,
            help='If none, the smallest dot has size 0. All expression levels with '
            'dot_min are potted with smallest_dot dot size.'
        ),
    ],

    'heat': [
        click.option(
            '--show-gene-labels',
            is_flag=True,
            default=None,
            show_default=True,
            help='By default gene labels are shown when there are 50 or less '
            'genes. Otherwise the labels are removed.'
        ),
    ],

    'swap_axes': click.option(
        '--swap-axes',
        is_flag=True,
        default=False,
        show_default=True,
        help='By default, the x axis contains var_names (e.g. genes) and the y '
        'axis the groupby categories. By setting swap_axes then x are the '
        'groupby categories and y the var_names. When swapping axes '
        'var_group_positions are no longer used.',
    ),

    'rank_genes_groups_plots': [
        click.option(
            '--groups',
            type=CommaSeparatedText(),
            default=None,
            show_default=True,
            help='The groups for which to show the gene ranking.'
        ),
        click.option(
            '--n-genes', '-n',
            type=click.INT,
            default=10,
            show_default=True,
            help='Number of genes to show.'
        ),
    ],

    'root': click.option(
        '--root',
        type=click.INT,
        default=0,
        show_default=True,
        help='If choosing a tree layout, this is the index of the root node.',
    ),

    'plot_embed': [
        click.option(
            '--use-raw/--no-raw',
            default=None,
            show_default=True,
            help='Use `.raw` attribute for coloring with gene expression. If '
            '`None`, uses `.raw` if present.',
        ),
        click.option(
            '--groups',
            type=click.STRING,
            default=None,
            help='Key for categorical in `.obs`. You can pass your predefined '
            'groups by choosing any categorical annotation of observations.',
        ),
    ],

    'batch_key': click.option(
        '--batch-key', 'key',
        type=click.STRING,
        required=True,
        help='The name of the column in adata.obs that differentiates among '
        'experiments/batches.'
    ),
    
    'batch_layer': click.option(
        '--layer', '-l',
        type=click.STRING,
        default=None,
        show_default=True,
        help="Layer to batch correct. By default corrects the contents of .X."
    ),
}

COMMON_OPTIONS['opt_output'] = [
    click.option(
        '--output-obj',
        type=click.Path(dir_okay=False, writable=True),
        help='Optionally output an object to the specified path.',
    ),
    *COMMON_OPTIONS['output'][1:], 
]

CMD_OPTIONS = {
    'read': [
        click.option(
            '--input-10x-h5', '-i',
            type=click.Path(exists=True, dir_okay=False),
            callback=mutually_exclusive_with('--input-10x-mtx'),
            help='Input 10x data in Cell-Ranger hdf5 format.',
        ),
        click.option(
            '--input-10x-mtx', '-x',
            type=click.Path(exists=True, file_okay=False),
            callback=mutually_exclusive_with('--input-10x-h5'),
            help='Path of input folder containing 10x data in mtx format.',
        ),
        *COMMON_OPTIONS['output'],
        click.option(
            '--genome', '-g',
            callback=required_by('--input-10x-h5'),
            default='hg19',
            show_default=True,
            help='Name of the genome group in hdf5 file, required by '
            '"--input-10x-h5".',
        ),
        click.option(
            '--var-names', '-v',
            type=click.Choice(['gene_symbols', 'gene_ids']),
            callback=required_by('--input-10x-mtx'),
            default='gene_symbols',
            show_default=True,
            help='Attribute to be used as the index of the variable table, '
            'required by "--input-10x-mtx".',
        ),
        click.option(
            '--extra-obs',
            type=click.Path(exists=True, dir_okay=False),
            default=None,
            show_default=True,
            help='Extra cell metadata table, must be tab-separated with a header '
            'row and an index column, and with matched dimension.',
        ),
        click.option(
            '--extra-var',
            type=click.Path(exists=True, dir_okay=False),
            default=None,
            show_default=True,
            help='Extra gene metadata table, must be tab-separated with a header '
            'row and an index column, and with matched dimension.',
        ),
    ],

    'filter': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['save'][0], # --save-raw
        click.option(
            '--gene-name', '-g',
            type=click.STRING,
            default='index',
            show_default=True,
            help='Name of the variable that contains gene names, used for flagging '
            'mitochondria genes when column "mito" is absent from `.var`.',
        ),
        click.option(
            '--list-attr', '-l',
            is_flag=True,
            default=False,
            help='When set, list attributes that can be filtered on.',
        ),
        click.option(
            '--param', '-p',
            type=(click.STRING, click.FLOAT, click.FLOAT),
            multiple=True,
            callback=valid_parameter_limits,
            help='Numerical parameters used to filter the data, '
            'in the format of "-p name min max". '
            'Multiple -p entries allowed.',
        ),
        click.option(
            '--category', '-c',
            type=(click.STRING, CommaSeparatedText()),
            multiple=True,
            help='Categorical attributes used to filter the data, '
            'in the format of "-c <name> <values>", '
            'where entries with attribute <name> with value in <values> are kept. '
            'If <values> is preceded by "!", entries with value in <values> are '
            'removed. Multiple -c entries allowed.',
        ),
        click.option(
            '--subset', '-s',
            type=(click.STRING, click.File()),
            multiple=True,
            help='Similar to --category in the format of "-s <name> <file>", '
            'but the <file> to be a one-column table that provides the values. '
            'Multiple -s entries allowed.',
        ),
        click.option(
            '--force-recalc',
            is_flag=True,
            default=False,
            help='When set, re-calculate `pct_counts_<qc_variable>` and '
            '`pct_counts_in_top_<n>_genes` even if they exist.',
        ),
    ],

    'norm': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        *COMMON_OPTIONS['save'],
        COMMON_OPTIONS['key_added'],
        click.option(
            '--no-log-transform', 'log_transform',
            is_flag=True,
            default=True,
            show_default=True,
            help='When set, do not apply (natural) log transform following normalisation.',
        ),
        click.option(
            '--normalize-to', '-t', 'target_sum',
            type=float,
            default=10_000,
            show_default=True,
            help='Normalize per cell nUMI to this number.',
        ),
        click.option(
            '--exclude-highly-expressed', '-e', 'exclude_highly_expressed',
            is_flag=True,
            default=False,
            show_default=True,
            help='Exclude (very) highly expressed genes for the computation of '
            'the normalization factor (size factor) for each cell. A gene is considered '
            'highly expressed, if it has more than max_fraction of the total counts in at '
            'least one cell. The not-excluded genes will sum up to the number '
            'specified by --normalize-to.'
        ),
        click.option(
            '--max-fraction', '-m', 'max_fraction',
            type=float,
            default=0.05,
            show_default=True,
            help='If exclude_highly_expressed=True, consider cells as highly '
            'expressed that have more counts than max_fraction of the original total counts '
            'in at least one cell.'
        ),
        click.option(
            '--layers', '-l',
            type=CommaSeparatedText(simplify=True),
            default=None,
            show_default=True,
            help="List of layers to normalize. Set to 'all' to normalize all layers."
        ),
        click.option(
            '--layer-norm', '-n', 'layer_norm',
            type=click.Choice(['after', 'X']),
            default=None,
            show_default=True,
            help="Specifies how to normalize layers: 1) If None, after "
            "normalization, for each layer in layers each cell has a total count equal to "
            "the median of the counts_per_cell before normalization of the layer. 2) If "
            "'after', for each layer in layers each cell has a total count equal to "
            "target_sum. 3) If 'X', for each layer in layers each cell has a total count "
            "equal to the median of total counts for observations (cells) of adata.X before "
            "normalization.'"
        ),
    ],

    'hvg': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        click.option(
            '--mean-limits', '-m',
            type=(click.FLOAT, click.FLOAT),
            callback=valid_limit,
            default=(0.0125, 3),
            show_default=True,
            help='Cutoffs for the mean of expression'
            'in the format of "-m min max".',
        ),
        click.option(
            '--disp-limits', '-d',
            type=(click.FLOAT, click.FLOAT),
            callback=valid_limit,
            default=(0.5, float('inf')),
            show_default=True,
            help='Cutoffs for the dispersion of expression'
            'in the format of "-d min max".',
        ),
        click.option(
            '--span',
            type=click.FLOAT,
            default=0.3,
            show_default=True,
            help="The fraction of the data (cells) used when estimating the "
            "variance in the loess model fit if flavor='seurat_v3'."    
        ),
        click.option(
            '--n-bins', '-b',
            type=click.INT,
            default=20,
            show_default=True,
            help='Number of bins for binning the mean gene expression.',
        ),
        click.option(
            '--n-top-genes', '-t',
            type=click.INT,
            default=None,
            show_default=True,
            help='Number of highly-variable genes to keep.',
        ),
        click.option(
            '--flavor', '-v',
            type=click.Choice(['seurat', 'cell_ranger', 'seurat_v3']),
            default='seurat',
            show_default=True,
            help='Choose the flavor for computing normalized dispersion.',
        ),
        click.option(
            '--subset', '-s',
            is_flag=True,
            default=False,
            help='When set, inplace subset to highly-variable genes, otherwise '
            'only flag highly-variable genes.',
        ),
        click.option(
            '--batch-key', 'batch_key',
            type=click.STRING,
            default=None,
            help="If specified, highly-variable genes are selected within each "
            "batch separately and merged. This simple process avoids the selection of "
            "batch-specific genes and acts as a lightweight batch correction method. For all "
            "flavors, genes are first sorted by how many batches they are a HVG. For "
            "dispersion-based flavors ties are broken by normalized dispersion. If flavor = "
            "'seurat_v3', ties are broken by the median (across batches) rank based on "
            "within-batch normalized variance."
        ),
    ],

    'scale': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        *COMMON_OPTIONS['save'],
        COMMON_OPTIONS['zero_center'],
        click.option(
            '--max-value', '-m',
            type=click.FLOAT,
            default=None,
            show_default=True,
            help='When specified, clip to this value after scaling, otherwise do '
            'not clip',
        ),
        click.option(
            '--layer', '-l',
            type=CommaSeparatedText(simplify=True),
            default=None,
            help="If provided, which element of layers to scale."
        ),
    ],

    'regress': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        *COMMON_OPTIONS['save'],
        COMMON_OPTIONS['n_jobs'],
        click.option(
            '--keys', '-k',
            type=CommaSeparatedText(simplify=True),
            default=None,
            show_default=True,
            help='Key(s) for observation annotation on which to regress.',
        ),
    ],

    'pca': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['zero_center'],
        COMMON_OPTIONS['random_state'],
        COMMON_OPTIONS['export_embedding'],
        COMMON_OPTIONS['n_comps'],
        click.option(
            '--svd-solver', '-V',
            type=click.Choice(['auto', 'arpack', 'randomized']),
            default='auto',
            show_default=True,
            help='SVD solver to use.'
        ),
        click.option(
            '--use-all', '-a', 'use_highly_variable',
            is_flag=True,
            flag_value=False,
            default=True,
            help='When set, use all genes for PCA, otherwise use '
            'highly-variable genes by default.'
        ),
        click.option(
            '--chunked', '-K',
            is_flag=True,
            default=False,
            help='When set, perform an incremental PCA on segments of '
            '--chunk-size, which automatically zero centers and ignore settings of '
            '--random-state and --svd-solver.',
        ),
        click.option(
            '--chunk-size', '-Z',
            type=click.INT,
            callback=required_by('--chunked'),
            default=None,
            show_default=True,
            help='Number of observations to include in each chunk, required by '
            '--chunked.',
        ),
    ],

    'neighbor': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        *COMMON_OPTIONS['use_pc'],
        COMMON_OPTIONS['key_added'],
        COMMON_OPTIONS['random_state'],
        click.option(
            '--n-neighbors', '-k',
            type=CommaSeparatedText(click.INT, simplify=True),
            default=15,
            show_default=True,
            help='The size of local neighborhood (in terms of number of '
            'neighboring data points) used for manifold approximation. Larger '
            'values result in more global views of the manifold, while smaller '
            'values result in more local data being preserved. In general values '
            'should be in the range 2 to 100.  If --knn is set, number of nearest '
            'neighbors to be searched, othwise a Gaussian kernel width is set to '
            'the distance of the --n-neighbors neighbor.',
        ),
        click.option(
            '--no-knn', 'knn',
            is_flag=True,
            flag_value=False,
            default=True,
            show_default=True,
            help='When NOT set, use a hard threshold to restrict the number of '
            'neighbors to --n-neighbors. Otherwise, use a Gaussian kernel to '
            'assign low weights to neighbors more distant than the --n-neighbors '
            'nearest neighbor',
        ),
        click.option(
            '--method', '-m',
            type=click.Choice(['umap', 'gauss', 'rapids']),
            default='umap',
            show_default=True,
            help='Use umap or gauss with adaptive width for computing '
            'connectivities. Use rapids for the RAPIDS implementation of UMAP '
            '(experimental, GPU only).'
        ),
        click.option(
            '--metric', '-t',
            type=click.Choice(['cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan', 'braycurtis', 'canberra', 'chebyshev', 'correlation', 'dice', 'hamming', 'jaccard', 'kulsinski', 'mahalanobis', 'minkowski', 'rogerstanimoto', 'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath', 'sqeuclidean', 'yule']),
            default='euclidean',
            show_default=True,
            help='A known metric’s name.'
        ),
    ],

    'umap': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['knn_graph'][0], # --neighbors-key
        COMMON_OPTIONS['random_state'],
        COMMON_OPTIONS['key_added'],
        COMMON_OPTIONS['export_embedding'],
        click.option(
            '--init-pos',
            type=click.STRING,
            default='spectral',
            show_default=True,
            help='How to initialize the low dimensional embedding. Can be '
            '"spectral", "paga" or "random", or any key of `.obsm`.',
        ),
        click.option(
            '--min-dist',
            type=click.FLOAT,
            default=0.5,
            show_default=True,
            help='The effective minimum distance between embedded points. Smaller '
            'values will result in a more clustered embedding, while larger values '
            'will results in a more even dispersal of points.',
        ),
        click.option(
            '--spread',
            type=click.FLOAT,
            default=1.0,
            show_default=True,
            help='The effective scale of embedded points, which determines the '
            'scale at which embedded points will be spread out.',
        ),
        click.option(
            '--n-components',
            type=click.INT,
            default=2,
            show_default=True,
            help='The number of dimensions of the embedding.',
        ),
        click.option(
            '--maxiter',
            type=click.INT,
            default=None,
            show_default=True,
            help='The number of iterations of the optimization.',
        ),
        click.option(
            '--alpha',
            type=click.FLOAT,
            default=1.0,
            show_default=True,
            help='The initial learning rate for the embedding optimization.',
        ),
        click.option(
            '--gamma',
            type=click.FLOAT,
            default=1.0,
            show_default=True,
            help='Weighting applied to negative samples in low dimensional '
            'embedding optimization.',
        ),
        click.option(
            '--negative-sample-rate',
            type=click.INT,
            default=5,
            show_default=True,
            help='The number of negative edge samples to use per positive edge '
            'sample in optimizing the low dimensional embedding.',
        ),
        click.option(
            '--method',
            type=click.Choice(['umap', 'rapids']),
            default='umap',
            show_default=True,
            help='Use the original ‘umap’ implementation, or ‘rapids’ '
            '(experimental, GPU only).'
        ),
    ],

    'tsne': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        *COMMON_OPTIONS['use_pc'],
        COMMON_OPTIONS['random_state'],
        COMMON_OPTIONS['key_added'],
        COMMON_OPTIONS['n_jobs'],
        COMMON_OPTIONS['export_embedding'],
        click.option(
            '--perplexity',
            type=click.FLOAT,
            default=30,
            show_default=True,
            help='The perplexity is related to the number of nearest neighbors '
            'that is used in other manifold learning algorithms. Larger datasets '
            'usually require a larger perplexity. Consider selecting a value '
            'between 5 and 50. The choice is not extremely critical since t-SNE '
            'is quite insensitive to this parameter.',
        ),
        click.option(
            '--early-exaggeration',
            type=click.FLOAT,
            default=12,
            show_default=True,
            help='Controls how tight natural clusters in the original space are in '
            'the embedded space and how much space will be between them. For '
            'larger values, the space between natural clusters will be larger in '
            'the embedded space. Again, the choice of this parameter is not very '
            'critical. If the cost function increases during initial optimization, '
            'the early exaggeration factor or the learning rate might be too high.',
        ),
        click.option(
            '--learning-rate',
            type=click.FLOAT,
            default=1000,
            show_default=True,
            help='Note that the R-package "Rtsne" uses a default of 200. The '
            'learning rate can be a critical parameter. It should be between 100 '
            'and 1000. If the cost function increases during initial optimization, '
            'the early exaggeration factor or the learning rate might be too high. '
            'If the cost function gets stuck in a bad local minimum increasing the '
            'learning rate helps sometimes.',
        ),
        click.option(
            '--no-fast-tsne', 'use_fast_tsne',
            is_flag=True,
            flag_value=False,
            default=True,
            show_default=True,
            help='When NOT set, use the MulticoreTSNE package by D. Ulyanov if '
            'installed.',
        ),
    ],

    'fdg': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['random_state'],
        COMMON_OPTIONS['export_embedding'],
        COMMON_OPTIONS['root'],
        click.option(
            '--init-pos',
            type=click.STRING,
            default=None,
            help='Use precomputed coordinates for initialization. Can be any key '
            'of `.obsm` or "paga" if .uns["paga"] is present',
        ),
        click.option(
            '--layout',
            type=click.Choice(['fa', 'fr', 'grid_fr', 'kk', 'lgl', 'drl', 'rt', 'rt_circular']),
            default='fa',
            show_default=True,
            help='Name of any valid igraph layout, including "fa" (ForceAtlas2), '
            '"fr" (Fruchterman Reingold), "grid_fr" (Grid Fruchterman Reingold, '
            'faster than "fr"), "kk" (Kamadi Kawai, slower than "fr"), "lgl" '
            '(Large Graph Layout, very fast), "drl" (Distributed Recursive Layout, '
            'pretty fast) and "rt" (Reingold Tilford tree layout).',
        ),
        click.option(
            '--key-added-ext',
            type=click.STRING,
            default=None,
            show_default=True,
            help="By default, append 'layout'"
        ),
        click.option(
            '--init-pos',
            type=click.STRING,
            default=None,
            show_default=True,
            help='How to initialize the low dimensional embedding. Can be "paga", '
            'or any valid key of `.obsm`.',
        ),
        COMMON_OPTIONS['knn_graph'][0], # --neighbors-key
        COMMON_OPTIONS['knn_graph'][1], # --obsp
    ],

    'louvain': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['export_cluster'],
        *COMMON_OPTIONS['knn_graph'],
        COMMON_OPTIONS['restrict_to'],
        COMMON_OPTIONS['random_state'],
        COMMON_OPTIONS['key_added'],
        click.option(
            '--flavor',
            type=click.Choice(['vtraag', 'igraph']),
            default='vtraag',
            show_default=True,
            help='Choose between two packages for computing the clustering. '
            '"vtraag" is much powerful, and the default.',
        ),
        click.option(
            '--resolution', '-r',
            type=CommaSeparatedText(click.FLOAT, simplify=True),
            default=1,
            show_default=True,
            help='For the default flavor "vtraag", you can provide a resolution. '
            'Higher resolution means finding more and smaller clusters.',
        ),
    ],

    'leiden': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['export_cluster'],
        *COMMON_OPTIONS['knn_graph'],
        COMMON_OPTIONS['restrict_to'],
        COMMON_OPTIONS['random_state'],
        COMMON_OPTIONS['key_added'],
        click.option(
            '--resolution', '-r',
            type=CommaSeparatedText(click.FLOAT, simplify=True),
            default=1,
            show_default=True,
            help='A parameter value controlling the coarseness of the clustering. '
            'Higher values lead to more clusters. Set to "None" if overriding '
            '--partition_type to one that doesn\'t accept `resolution_parameter`.',
        ),
        click.option(
            '--n-iterations',
            type=click.INT,
            default=-1,
            show_default=True,
            help='How many iterations of the Leiden clustering algorithm to '
            'perform. -1 has the algorithm run until it reaches its optimal '
            'clustering.',
        ),
    ],

    'diffexp': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['use_raw'],
        COMMON_OPTIONS['key_added'],
        click.option(
            '--layer', '-l',
            type=click.STRING,
            default=None,
            help="Key from adata.layers whose value will be used to perform tests on."
        ),
        click.option(
            '--groupby', '-g',
            type=click.STRING,
            required=True,
            help='The key of the observations grouping to consider.',
        ),
        click.option(
            '--groups',
            type=CommaSeparatedText(simplify=True),
            default='all',
            show_default=True,
            help='Subset of groups to which comparison shall be restricted.',
        ),
        click.option(
            '--reference',
            type=click.STRING,
            default='rest',
            show_default=True,
            help='If "rest", compare each group to the union of the rest of the '
            'groups. If a group identifier, compare with respect to this group.',
        ),
        click.option(
            '--n-genes', '-n',
            type=click.INT,
            default=None,
            show_default=True,
            help='The number of genes that appear in the retured tables. By '
            'default return all available genes depending on the value of '
            '--use-raw.'
        ),
        click.option(
            '--method',
            type=click.Choice(
                ['logreg', 't-test', 'wilcoxon', 't-test_overestim_var']),
            default='t-test_overestim_var',
            show_default=True,
            help='Method of performing differential expression analysis.',
        ),
        click.option(
            '--corr-method',
            type=click.Choice(['benjamini-hochberg', 'bonferroni']),
            default='benjamini-hochberg',
            show_default=True,
            help='P-value correction method. Used only for "t-test", '
            '"t-test_overestim_var" and "wilcoxon".',
        ),
        click.option(
            '--rankby-abs',
            is_flag=True,
            default=False,
            show_default=True,
            help='Rank genes by the absolute value of the score, not by the score. '
            'The returned scores are never the absolute values.',
        ),
        click.option(
            '--pts',
            is_flag=True,
            default=False,
            show_default=True,
            help='Compute the fraction of cells expressing the genes.'
        ),
        click.option(
            '--tie-correct',
            is_flag=True,
            default=False,
            show_default=True,
            help="Use tie correction for 'wilcoxon' scores. Used only for "
            "'wilcoxon'."
        ),
        click.option(
            '--filter-params',
            type=Dictionary(keys=[
                'min_in_group_fraction',
                'max_out_group_fraction',
                'min_fold_change',
            ]),
            default=None,
            show_default=True,
            help='Parameters for filtering DE results, valid parameters are: '
            '"min_in_group_fraction" (float), "max_out_group_fraction" (float), '
            '"min_fold_change" (float).',
        ),
        click.option(
            '--logreg-param',
            type=Dictionary(),
            default=None,
            show_default=True,
            help='Parameters passed to `sklearn.linear_model.LogisticRegression`.',
        ),
        click.option(
            '--save',
            type=click.Path(dir_okay=False, writable=True),
            default=None,
            show_default=True,
            help='Tab-separated table to store results of differential expression '
            'analysis.',
        ),
    ],

    'paga': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['knn_graph'][0], # --neighbors-key
        COMMON_OPTIONS['key_added'],
        click.option(
            '--groups',
            type=click.STRING,
            required=True,
            help='Key for categorical in `.obs`. You can pass your predefined '
            'groups by choosing any categorical annotation of observations.',
        ),
        click.option(
            '--model',
            type=click.Choice(['v1.2', 'v1.0']),
            default='v1.2',
            show_default=True,
            help='The PAGA connectivity model.',
        ),
        click.option(
            '--use-rna-velocity',
            is_flag=True,
            default=False,
            show_default=True,
            help='Use RNA velocity to orient edges in the abstracted graph and '
            'estimate transitions. Requires that adata.uns contains a directed single-cell '
            'graph with key velocity_graph. This feature might be subject to change in the '
            'future.',
        ),
    ],

    'diffmap': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['knn_graph'][0], # --neighbors-key
        COMMON_OPTIONS['key_added'],
        COMMON_OPTIONS['export_embedding'],
        COMMON_OPTIONS['n_comps'],
    ],

    'dpt': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['knn_graph'][0], # --neighbors-key
        COMMON_OPTIONS['key_added'],
        click.option(
            '--root',
            type=(click.STRING, click.STRING),
            default=(None, None),
            show_default=True,
            help='Specify a categorical annotaion of observations (`.obs`) and a '
            'value representing the root cells.',
        ),
        click.option(
            '--n-dcs',
            type=click.INT,
            default=10,
            show_default=True,
            help='The number of diffusion components to use.',
        ),
        click.option(
            '--n-branchings',
            type=click.INT,
            default=0,
            show_default=True,
            help='Number of branchings to detect.',
        ),
        click.option(
            '--min-group-size',
            type=click.FLOAT,
            default=0.01,
            show_default=True,
            help='During recursive splitting of branches for --n-branchings > 1, '
            'do not consider branches/groups that contain fewer than this fraction '
            'of the total number of data points.',
        ),
        click.option(
            '--disallow-kendall-tau-shift', 'allow_kendall_tau_shift',
            is_flag=True,
            default=True,
            show_default=True,
            help='By default: If a very small branch is detected upon '
            'splitting, shift away from maximum correlation in Kendall tau criterion of '
            '[Haghverdi16] to stabilize the splitting. Use flag to disable this.'
        ),
    ],

    'combat': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['batch_key'],
        COMMON_OPTIONS['batch_layer'],
        click.option(
            '--key-added',
            type=click.STRING,
            default=None,
            show_default=True,
            help="Key under which to add the computed results. By default a new "
            "layer will be created called 'combat', 'combat_{layer}' or "
            "'combat_layer_{key_added}' where those parameters were specified. A value of 'X' "
            "causes batch-corrected values to overwrite the original content of .X."
        ),
        click.option(
            '--covariates',
            type=(CommaSeparatedText()),
            default=None,
            show_default=True,
            help="Comma-separated list of additional covariates besides the "
            "batch variable such as adjustment variables or biological condition. This "
            "parameter refers to the design matrix X in Equation 2.1 in [Johnson07] and to "
            "the mod argument in the original combat function in the sva R package.  Note "
            "that not including covariates may introduce bias or lead to the removal of "
            "biological signal in unbalanced designs."
        ),
        
    ],

    'harmony': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['batch_key'],
        click.option(
            '--basis',
            type=click.STRING,
            default='X_pca',
            show_default=True,
            help="The name of the field in adata.obsm where the PCA table is "
            "stored. Defaults to 'X_pca', which is the default for sc.tl.pca()."
        ),
        click.option(
            '--adjusted-basis',
            type=click.STRING,
            default='X_pca_harmony',
            show_default=True,
            help='The name of the field in adata.obsm where the adjusted PCA '
            'table will be stored after running this function.'
        ),
        click.option(
            '--theta',
            type=click.FLOAT,
            default=2,
            show_default=True,
            help='Diversity clustering penalty parameter. theta=0 does not encourage any '
            'diversity. Larger values of theta result in more diverse clusters.'
        ),
        click.option(
            '--lambda', 'lamb',
            type=click.FLOAT,
            default=1,
            show_default=True,
            help='Ridge regression penalty parameter. Lambda must be strictly '
            'positive.  Smaller values result in more aggressive correction.'
        ),
        click.option(
            '--sigma',
            type=click.FLOAT,
            default=0.1,
            show_default=True,
            help='Width of soft kmeans clusters. Sigma scales the distance from '
            'a cell to cluster centroids. Larger values of sigma result in cells assigned to '
            'more clusters. Smaller values of sigma make soft kmeans cluster approach hard '
            'clustering.'
        ),
        click.option(
            '--n-clust', 'nclust',
            type=click.INT,
            default=None,
            show_default=False,
            help='Number of clusters in model. nclust=1 equivalent to simple '
            'linear regression.'
        ),
        click.option(
            '--tau',
            type=click.INT,
            default=0,
            show_default=True,
            help='Protection against overclustering small datasets with large ones. '
            'tau is the expected number of cells per cluster.'
        ),
        click.option(
            '--block-size',
            type=click.FLOAT,
            default=0.05,
            show_default=True,
            help='What proportion of cells to update during clustering. Between '
            '0 to 1, default 0.05. Larger values may be faster but less accurate.'
        ),
        click.option(
            '--max-iter-cluster', 'max_iter_kmeans',
            type=click.INT,
            default=20,
            show_default=True,
            help='Maximum number of rounds to run clustering at each round of '
            'Harmony.'
        ),
        click.option(
            '--max-iter-harmony',
            type=click.INT,
            default=10,
            show_default=True,
            help='Maximum number of rounds to run Harmony. One round of Harmony '
            'involves one clustering and one correction step.'
        ),
        click.option(
            '--epsilon-cluster',
            type=click.FLOAT,
            default=1e-5,
            show_default=True,
            help='Convergence tolerance for clustering round of Harmony Set to '
            '-Inf to never stop early.'
        ),
        click.option(
            '--epsilon-harmony',
            type=click.FLOAT,
            default=1e-5,
            show_default=True,
            help='Convergence tolerance for clustering round of Harmony Set to '
            '-Inf to never stop early.'
        ),
        COMMON_OPTIONS['random_state'],
    ],

    'mnn': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        *COMMON_OPTIONS['save'],
        COMMON_OPTIONS['batch_key'],
        COMMON_OPTIONS['batch_layer'],
        click.option(
            '--key-added',
            type=click.STRING,
            default=None,
            show_default=True,
            help="Key under which to add the computed results. By default a new "
            "layer will be created called 'mnn', 'mnn_{layer}' or "
            "'mnn_layer_{key_added}' where those parameters were specified. A value of 'X' "
            "causes batch-corrected values to overwrite the original content of .X."
        ),
        click.option(
            '--var-subset',
            type=(click.STRING, CommaSeparatedText()),
            multiple=True,
            help="The subset of vars (list of str) to be used when performing "
            "MNN correction in the format of '--var-subset <name> <values>'. Typically, use "
            "the highly variable genes (HVGs) like '--var-subset highly_variable True'. When "
            "unset, uses all vars."
        ),
        click.option(
            '--n-neighbors', '-k',
            type=CommaSeparatedText(click.INT, simplify=True),
            default=20,
            show_default=True,
            help='Number of mutual nearest neighbors.'
        ),
        click.option(
            '--sigma',
            type=click.FLOAT,
            default=1.0,
            show_default=True,
            help='The bandwidth of the Gaussian smoothing kernel used to '
            'compute the correction vectors.'
        ),
        click.option(
            '--no-cos_norm_in', 'cos_norm_in',
            is_flag=True,
            default=True,
            help='Default behaviour is to perform cosine normalization on the '
            'input data prior to calculating distances between cells. Use this '
            'flag to disable that behaviour.'
        ),
        click.option(
            '--no-cos_norm_out', 'cos_norm_out',
            is_flag=True,
            default=True,
            help='Default behaviour is to perform cosine normalization prior to '
            'computing corrected expression values. Use this flag to disable that '
            'behaviour.'
        ),
        click.option(
            '--svd-dim',
            type=click.INT,
            default=None,
            show_default=True,
            help='The number of dimensions to use for summarizing biological '
            'substructure within each batch. If not set, biological components '
            'will not be removed from the correction vectors.'
        ),
        click.option(
            '--no-var-adj',
            is_flag=True,
            default=True,
            help='Default behaviour is to adjust variance of the correction '
            'vectors. Use this flag to disable that behaviour. Note this step takes most '
            'computing time.'
        ),
        click.option(
            '--compute-angle',
            is_flag=True,
            default=False,
            help='When set, compute the angle between each cell’s correction '
            'vector and the biological subspace of the reference batch.'
        ),
        click.option(
            '--svd-mode',
            type=click.Choice(['svd', 'rsvd', 'irlb']),
            default='rsvd',
            show_default=True,
            help="'svd' computes SVD using a non-randomized SVD-via-ID "
            "algorithm, while 'rsvd' uses a randomized version. 'irlb' performs truncated "
            "SVD by implicitly restarted Lanczos bidiagonalization (forked from "
            "https://github.com/airysen/irlbpy)."
        ),
    ],

    'bbknn': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['output'],
        COMMON_OPTIONS['key_added'],
        COMMON_OPTIONS['batch_key'],
        click.option(
            '--use-rep', '-u',
            type=click.STRING,
            default='X_pca',
            show_default=True,
            help='The dimensionality reduction in .obsm to use for neighbour '
            'detection.'
        ),
        COMMON_OPTIONS['use_pc'][0], # --n-pcs
        click.option(
            '--no-approx', 'approx',
            is_flag=True,
            default=True,
            help='Default behaviour is to use annoy’s approximate neighbour '
            'finding. This results in a quicker run time for large datasets while also '
            'potentially increasing the degree of batch correction. Use this flag to disable '
            'that behaviour.',
        ),
        click.option(
            '--metric', '-t',
            type=click.Choice(['angular', 'cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan', 'braycurtis', 'canberra', 'chebyshev', 'correlation', 'dice', 'hamming', 'jaccard', 'kulsinski', 'mahalanobis', 'minkowski', 'rogerstanimoto', 'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath', 'sqeuclidean', 'yule']),
            default='angular',
            show_default=True,
            help='A known metric’s name.'
        ),
        click.option(
            '--neighbors-within-batch',
            type=click.INT,
            default=3,
            show_default=True,
            help='How many top neighbours to report for each batch; total '
            'number of neighbours will be this number times the number of batches.'
        ),
        click.option(
            '--trim',
            type=click.INT,
            default=None,
            show_default=True,
            help='Trim the neighbours of each cell to these many top '
            'connectivities. May help with population independence and improve the tidiness '
            'of clustering. The lower the value the more independent the individual '
            'populations, at the cost of more conserved batch effect. If None, sets the '
            'parameter value automatically to 10 times the total number of neighbours for '
            'each cell. Set to 0 to skip.'
        ),
        click.option(
            '--n-trees',
            type=click.INT,
            default=10,
            show_default=True,
            help='Only used when approx=True. The number of trees to construct '
            'in the annoy forest. More trees give higher precision when querying, at the '
            'cost of increased run time and resource intensity.'
        ),
        click.option(
            '--no-use-faiss', 'use_faiss',
            is_flag=True,
            default=True,
            help='Default behaviour If approx=False and the metric is '
            '“euclidean”, is to use the faiss package to compute nearest neighbours if '
            'installed. This improves performance at a minor cost to numerical precision as '
            'faiss operates on float32. Use this flag to disable that behaviour.'
        ),
        click.option(
            '--set-op-mix-ratio',
            type=click.FLOAT,
            default=1,
            show_default=True,
            help='UMAP connectivity computation parameter, float between 0 and '
            '1, controlling the blend between a connectivity matrix formed exclusively from '
            'mutual nearest neighbour pairs (0) and a union of all observed neighbour '
            'relationships with the mutual pairs emphasised (1).'
        ),
        click.option(
            '--local-connectivity',
            type=click.INT,
            default=1,
            show_default=True,
            help='UMAP connectivity computation parameter, how many nearest '
            'neighbors of each cell are assumed to be fully connected (and given a '
            'connectivity value of 1)'
        ),
    ],

    'embed': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['plot'],
        *COMMON_OPTIONS['frame_title'],
        COMMON_OPTIONS['layer'],
        click.option(
            '--basis',
            type=click.STRING,
            default='umap',
            show_default=True,
            help='Name of the embedding to plot, must be a key of `.obsm` without '
            'the prefix "X_".',
        ),
        click.option(
            '--color',
            type=CommaSeparatedText(simplify=True),
            default=None,
            show_default=True,
            help='Keys for annotations of observations/cells or variables/genes.',
        ),
        click.option(
            '--legend-loc',
            type=click.Choice(['right margin', 'on data']),
            default='right margin',
            show_default=True,
            help='Location of legend, either "on data", "right margin" or valid '
            'keywords for `matplotlib.legend`.',
        ),
        click.option(
            '--legend-fontsize',
            type=click.INT,
            default=15,
            show_default=True,
            help='Legend font size.',
        ),
        click.option(
            '--size',
            type=click.FLOAT,
            default=None,
            show_default=True,
            help='Point size. Automatically computed if not specified.',
        ),
        COMMON_OPTIONS['gene_symbols'],
        click.option(
            '--edges',
            is_flag=True,
            default=False,
            show_default=True,
            help='Show edges.',
        ),
        click.option(
            '--edges-width',
            type=click.FLOAT,
            default=0.1,
            show_default=True,
            help='Width of edges.',
        ),
        click.option(
            '--edges-color',
            type=click.STRING,
            default=None,
            show_default=True,
            help='Color of edges. See draw_networkx_edges().',
        ),
        COMMON_OPTIONS['knn_graph'][0], # --neighbors-key
        click.option(
            '--no-sort-order', 'sort_order',
            is_flag=True,
            default=True,
            show_default=True,
            help='Disable default behaviour: for continuous annotations used as '
            'color parameter, plot data points with higher values on top of others.',
        ),
        *COMMON_OPTIONS['plot_embed'],
        click.option(
            '--components',
            type=click.STRING,
            default=None,
            show_default=True,
            help="For instance, ['1,2', '2,3']. To plot all available components use 'all'.",
        ),
        click.option(
            '--projection',
            type=click.Choice(['2d', '3d']),
            default='2d',
            show_default=True,
            help="Projection of plot."
        ),
         
    ],

    'plot_paga': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['plot'],
        *COMMON_OPTIONS['frame_title'],
        *COMMON_OPTIONS['plot_embed'],
        COMMON_OPTIONS['random_state'],
        click.option(
            '--use-key',
            type=click.STRING,
            default='paga',
            show_default=True,
            help='The key in `.uns` that contains trajectory information.',
        ),
        click.option(
            '--layout',
            type=click.Choice(['fa', 'fr', 'grid_fr', 'kk', 'lgl', 'drl', 'rt']),
            default='fr',
            show_default=True,
            help='Plotting layout that computes positions.',
        ),
        click.option(
            '--init-pos',
            type=click.STRING,
            default=None,
            show_default=True,
            help='Plotting layout that computes positions.',
        ),
        click.option(
            '--threshold',
            type=click.FLOAT,
            default=0.01,
            show_default=True,
            help='Do not draw edges for weights below this threshold. Set to 0 to '
            'include all edges.',
        ),
        COMMON_OPTIONS['root'],
        click.option(
            '--root',
            type=click.INT,
            default=0,
            show_default=True,
            help='If choosing a tree layout, this is the index of the root node.',
        ),
        click.option(
            '--transitions',
            type=click.STRING,
            default=None,
            show_default=True,
            help='Key for `.uns["paga"]` that specifies the matrix, e.g. '
            '`transition_confidence`, that stores the arrows.',
        ),
        click.option(
            '--single-component',
            is_flag=True,
            default=False,
            show_default=True,
            help='Restrict to largest connected component',
        ),
        click.option(
            '--solid-edges',
            type=click.Choice(['connectivities', 'connectivities_tree']),
            default='connectivities',
            show_default=True,
            help='Key for `.uns["paga"]` that specifies the matrix that stores the '
            'edges to be drawn solid black.',
        ),
        click.option(
            '--basis',
            type=click.STRING,
            default=None,
            show_default=True,
            help='Name of the embedding to plot, must be a key of `.obsm` without '
            'the prefix "X_".',
        ),
        click.option(
            '--color',
            type=click.STRING,
            default=None,
            show_default=True,
            help='Key for annotation of observations/cells or variables/genes.',
        ),
        click.option(
            '--legend-loc',
            type=click.Choice(['right margin', 'on data']),
            default='right margin',
            show_default=True,
            help='Location of legend, either "on data", "right margin" or valid '
            'keywords for `matplotlib.legend`.',
        ),
        click.option(
            '--size',
            type=click.FLOAT,
            default=None,
            show_default=True,
            help='Point size. Automatically computed if not specified.',
        ),
        click.option(
            '--node-size-scale',
            type=click.FLOAT,
            default=1.0,
            show_default=True,
            help='Increase of decrease the size of the nodes.',
        ),
        click.option(
            '--fontsize',
            type=click.INT,
            default=None,
            show_default=True,
            help='Font size for node labels.',
        ),
        click.option(
            '--edge-width-scale',
            type=click.FLOAT,
            default=1.0,
            show_default=True,
            help='Increase of decrease the width of the edges.',
        ),
        click.option(
            '--arrowsize',
            type=click.INT,
            default=30,
            show_default=True,
            help='For directed graphs, specify the length and width of the arrowhead.',
        ),
        *COMMON_OPTIONS['opt_output'],
    ],

    'sviol': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['plot'],
        COMMON_OPTIONS['use_raw'],
        COMMON_OPTIONS['var_names'],
        *COMMON_OPTIONS['rank_genes_groups_plots'],
        COMMON_OPTIONS['layer'],
        *COMMON_OPTIONS['diffexp_plot'],
        COMMON_OPTIONS['gene_symbols'],
        *COMMON_OPTIONS['sviol'],
        COMMON_OPTIONS['swap_axes'],
    ],

    'dot': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['plot'],
        COMMON_OPTIONS['use_raw'],
        COMMON_OPTIONS['var_names'],
        *COMMON_OPTIONS['rank_genes_groups_plots'],
        COMMON_OPTIONS['layer'],
        *COMMON_OPTIONS['diffexp_plot'],
        COMMON_OPTIONS['gene_symbols'],
        *COMMON_OPTIONS['dot'],
    ],

    'matrix': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['plot'],
        COMMON_OPTIONS['use_raw'],
        COMMON_OPTIONS['var_names'],
        *COMMON_OPTIONS['rank_genes_groups_plots'],
        COMMON_OPTIONS['layer'],
        *COMMON_OPTIONS['diffexp_plot'],
        COMMON_OPTIONS['gene_symbols'],
    ],

    'heat': [
        *COMMON_OPTIONS['input'],
        *COMMON_OPTIONS['plot'],
        COMMON_OPTIONS['use_raw'],
        COMMON_OPTIONS['var_names'],
        *COMMON_OPTIONS['rank_genes_groups_plots'],
        COMMON_OPTIONS['layer'],
        *COMMON_OPTIONS['diffexp_plot'],
        COMMON_OPTIONS['gene_symbols'],
        *COMMON_OPTIONS['heat'],
        COMMON_OPTIONS['swap_axes'],
    ],

}



"""
scanpy pca
"""

import logging
import scanpy as sc
from ..obj_utils import write_embedding

def pca(adata, key_added=None, export_embedding=None, **kwargs):
    """
    Wrapper function for sc.pp.pca, for supporting named slot
    """

    # omit "svd_solver" to let scanpy choose automatically
    if 'svd_solver' in kwargs and kwargs['svd_solver'] == 'auto':
        del kwargs['svd_solver']

    if key_added:
        if 'X_pca' in adata.obsm.keys():
            adata.obsm['X_pca_bkup'] = adata.obsm['X_pca']
        sc.pp.pca(adata, **kwargs)
        pca_key = f'X_pca_{key_added}'
        adata.obsm[pca_key] = adata.obsm['X_pca']
        del adata.obsm['X_pca']
        if 'X_pca_bkup' in adata.obsm.keys():
            adata.obsm['X_pca'] = adata.obsm['X_pca_bkup']
            del adata.obsm['X_pca_bkup']
    else:
        sc.pp.pca(adata, **kwargs)
        pca_key = 'X_pca'

    if export_embedding is not None:
        write_embedding(adata, pca_key, export_embedding, key_added=key_added)
    return adata



"""
Provides version, author and exports
"""
import pkg_resources

__version__ = pkg_resources.get_distribution('scanpy-scripts').version

__author__ = ', '.join([
    'Ni Huang',
    'Pablo Moreno',
    'Jonathan Manning',
    'Philipp Angerer',
])

from . import lib



"""
scanpy combat
"""

import scanpy as sc

# Wrapper for mnn allowing use of non-standard slot

def combat(adata, key=None, key_added=None, layer=None, **kwargs):
    """
    Wrapper function for scanpy.pp.combat(), for supporting non-standard slots
    """

    # If layer is set then we have to move the contents of that layer into
    # .X for analysis. We back up the original .X, but only if the user hasn't
    # specified to overwrite it anyway.

    if layer:
        if key_added and key_added != 'X':
            adata.layers['X_backup'] = adata.X

        adata.X = adata.layers[layer]
    
    # If we're storing results in .X (whether from .X or from a layer), run in
    # place to save copying objects.

    if key_added and key_added == 'X':
        sc.pp.combat(adata, key = key, **kwargs)    
    
    # If we're storing in 'layers' (key_added is not set, or is not X, then
    # don't run in place, and put the matrix in the specified layer. 

    else:

        cdata = sc.pp.combat(adata, key=key, inplace = False, **kwargs)    
            
        combat_key = 'combat'
        if layer:
            combat_key = f"{combat_key}_{layer}"
        
            # If we ran from a layer, restore the .X we had to overwrite
            
            adata.X = adata.layers['X_backup']    
            del adata.layers['X_backup'] 
            
        if key_added:
            combat_key = f"{combat_key}_{key_added}"

        adata.layers[combat_key] = cdata

    return adata



"""
scanpy
"""

import logging
import click
import scanpy as sc
from .click_utils import NaturalOrderGroup
from .cmds import (
    READ_CMD,
    FILTER_CMD,
    NORM_CMD,
    HVG_CMD,
    SCALE_CMD,
    REGRESS_CMD,
    PCA_CMD,
    NEIGHBOR_CMD,
    UMAP_CMD,
    TSNE_CMD,
    FDG_CMD,
    LOUVAIN_CMD,
    LEIDEN_CMD,
    DIFFEXP_CMD,
    PAGA_CMD,
    DIFFMAP_CMD,
    DPT_CMD,
    PLOT_EMBED_CMD,
    PLOT_PAGA_CMD,
    PLOT_STACKED_VIOLIN_CMD,
    PLOT_DOT_CMD,
    PLOT_MATRIX_CMD,
    PLOT_HEATMAP_CMD,
    HARMONY_INTEGRATE_CMD,
    BBKNN_CMD,
    MNN_CORRECT_CMD,
    COMBAT_CMD,
)


@click.group(cls=NaturalOrderGroup)
@click.option(
    '--debug',
    is_flag=True,
    default=False,
    help='Print debug information',
)
@click.option(
    '--verbosity',
    type=click.INT,
    default=3,
    help='Set scanpy verbosity',
)
@click.version_option(
    version='0.2.0',
    prog_name='scanpy',
)
def cli(debug=False, verbosity=3):
    """
    Command line interface to [scanpy](https://github.com/theislab/scanpy)
    """
    log_level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(
        level=log_level,
        format=('%(asctime)s; %(levelname)s; %(filename)s; '
                '%(funcName)s(): %(message)s'),
        datefmt='%y-%m-%d %H:%M:%S',
    )
    logging.debug('debugging')
    sc.settings.verbosity = verbosity
    return 0


cli.add_command(READ_CMD)
cli.add_command(FILTER_CMD)
cli.add_command(NORM_CMD)
cli.add_command(HVG_CMD)
cli.add_command(SCALE_CMD)
cli.add_command(REGRESS_CMD)
cli.add_command(PCA_CMD)
cli.add_command(NEIGHBOR_CMD)


@cli.group(cls=NaturalOrderGroup)
def embed():
    """Embed cells into two-dimensional space."""


embed.add_command(UMAP_CMD)
embed.add_command(TSNE_CMD)
embed.add_command(FDG_CMD)
embed.add_command(DIFFMAP_CMD)


@cli.group(cls=NaturalOrderGroup)
def cluster():
    """Cluster cells into sub-populations."""


cluster.add_command(LOUVAIN_CMD)
cluster.add_command(LEIDEN_CMD)


cli.add_command(DIFFEXP_CMD)
cli.add_command(PAGA_CMD)
cli.add_command(DPT_CMD)


@cli.group(cls=NaturalOrderGroup)
def integrate():
    """Integrate cells from different experimental batches."""

integrate.add_command(HARMONY_INTEGRATE_CMD)
integrate.add_command(BBKNN_CMD)
integrate.add_command(MNN_CORRECT_CMD)
integrate.add_command(COMBAT_CMD)


@cli.group(cls=NaturalOrderGroup)
def plot():
    """Visualise data."""


plot.add_command(PLOT_EMBED_CMD)
plot.add_command(PLOT_PAGA_CMD)
plot.add_command(PLOT_STACKED_VIOLIN_CMD)
plot.add_command(PLOT_DOT_CMD)
plot.add_command(PLOT_MATRIX_CMD)
plot.add_command(PLOT_HEATMAP_CMD)



"""
scanpy louvain
"""

import scanpy as sc
from ..obj_utils import write_cluster


def louvain(
        adata,
        resolution,
        neighbors_key=None,
        obsp=None,
        key_added=None,
        export_cluster=None,
        **kwargs
):
    """
    Wrapper function for sc.tl.louvain, for supporting multiple resolutions.
    """
    keys = []
    if kwargs['restrict_to'] and not kwargs['restrict_to'][0]:
        kwargs['restrict_to'] = None
    
    if not isinstance(resolution, (list, tuple)):
        if key_added is not None and not key_added.startswith('louvain_'):
            key_added = f'louvain_{key_added}'
        elif key_added is None:
            key_added = 'louvain'
        sc.tl.louvain(
            adata,
            resolution=resolution,
            key_added=key_added,
            neighbors_key=neighbors_key,
            obsp=obsp,
            **kwargs
        )
        keys.append(key_added)
    else:
        for i, res in enumerate(resolution):

            res_key = str(res).replace('.', '_')

            if key_added is None:
                graph_key = ('_' + f'{neighbors_key or obsp}') if neighbors or obsp else ''
                key = f'louvain{graph_key}_r{res_key}'    
            elif not isinstance(key_added, (list, tuple)):
                key = f'louvain_{key_added}_r{res_key}'
            elif len(key_added) == len(resolution):
                key = key_added[i]
            else:
                raise ValueError('`key_added` can only be None, a scalar, or an '
                                 'iterable of the same length as `resolution`.')
            keys.extend(louvain(
                adata,
                resolution=res,
                neighbors_key=neighbors_key,
                obsp=obsp,
                key_added=key,
                **kwargs,
            ))

    if export_cluster:
        write_cluster(adata, keys, export_cluster)

    return keys



"""
scanpy filter
"""

import logging
import re
import click
import numpy as np
import scanpy as sc


def filter_anndata(
        adata,
        gene_name='index',
        list_attr=False,
        param=None,
        category=None,
        subset=None,
        force_recalc=False,
):
    """
    Wrapper function for sc.pp.filter_cells() and sc.pp.filter_genes(), mainly
    for supporting arbitrary filtering
    """
    param = [] if param is None else param
    category = [] if category is None else category
    subset = [] if subset is None else subset

    logging.debug('--gene-name=%s', gene_name)
    logging.debug('--param=%s', param)
    logging.debug('--category=%s', category)
    logging.debug('--subset=%s', subset)

    if 'mito' not in adata.var.keys() and gene_name:
        try:
            gene_names = getattr(adata.var, gene_name)
            k_mito = gene_names.str.startswith('MT-')
            if k_mito.sum() > 0:
                adata.var['mito'] = k_mito
                adata.var['mito'] = adata.var['mito'].astype('category')
            else:
                logging.warning('No MT genes found, skip calculating '
                                'expression of mitochondria genes')
        except AttributeError:
            logging.warning(
                'Specified gene column [%s] not found, skip calculating '
                'expression of mitochondria genes', gene_name)

    attributes = _get_attributes(adata)
    if list_attr:
        click.echo(_repr_obj(attributes))
        return 0

    conditions, qc_vars, pct_top = _get_filter_conditions(
        attributes, param, category, subset)

    layer = 'counts' if 'counts' in adata.layers.keys() else None
    obs_columns = adata.obs.columns
    for qv in qc_vars:
        if f'pct_counts_{qv}' in obs_columns and not force_recalc:
            logging.warning('`pct_counts_%s` exists, not overwriting '
                            'without --force-recalc', qv)
            qc_vars.remove(qv)
    for pt in pct_top:
        if f'pct_counts_in_top_{pt}_genes' in obs_columns and not force_recalc:
            logging.warning('`pct_counts_%s` exists, not overwriting '
                            'without --force-recalc', pt)
            pct_top.remove(pt)
    sc.pp.calculate_qc_metrics(
        adata, layer=layer, qc_vars=qc_vars, percent_top=pct_top, inplace=True)
    adata.obs['n_counts'] = adata.obs['total_counts']
    adata.obs['n_genes'] = adata.obs['n_genes_by_counts']
    adata.var['n_counts'] = adata.var['total_counts']
    adata.var['n_cells'] = adata.var['n_cells_by_counts']

    k_cell = np.ones(len(adata.obs)).astype(bool)
    for cond in conditions['c']['numerical']:
        name, vmin, vmax = cond
        attr = adata.obs[name]
        k_cell = k_cell & (attr >= vmin) & (attr <= vmax)

    for cond in conditions['c']['categorical']:
        name, values = cond
        attr = getattr(adata.obs, name).astype(str)
        if values[0].startswith('!'):
            values[0] = values[0][1:]
            k_cell = k_cell & (~attr.isin(values))
        else:
            k_cell = k_cell & attr.isin(values)

    k_gene = np.ones(len(adata.var)).astype(bool)
    for cond in conditions['g']['numerical']:
        name, vmin, vmax = cond
        attr = adata.var[name]
        k_gene = k_gene & (attr >= vmin) & (attr <= vmax)

    for cond in conditions['g']['categorical']:
        name, values = cond
        attr = getattr(adata.var, name).astype(str)
        if values[0].startswith('!'):
            values[0] = values[0][1:]
            k_gene = k_gene & ~(attr.isin(values))
        else:
            k_gene = k_gene & attr.isin(values)

    adata._inplace_subset_obs(k_cell)
    adata._inplace_subset_var(k_gene)

    return adata


def _get_attributes(adata):
    attributes = {
        'c': {
            'numerical': [],
            'categorical': ['index'],
            'bool': [],
        },
        'g': {
            'numerical': [],
            'categorical': ['index'],
            'bool': [],
        },
    }

    for attr, dtype in adata.obs.dtypes.to_dict().items():
        typ = dtype.kind
        if typ == 'O':
            if dtype.name == 'category' and dtype.categories.is_boolean():
                attributes['c']['bool'].append(attr)
            attributes['c']['categorical'].append(attr)
        elif typ in ('i', 'f', 'u'):
            attributes['c']['numerical'].append(attr)
        elif typ == 'b':
            attributes['c']['bool'].append(attr)
            attributes['c']['categorical'].append(attr)

    for attr, dtype in adata.var.dtypes.to_dict().items():
        typ = dtype.kind
        if typ == 'O':
            if dtype.name == 'category' and dtype.categories.is_boolean():
                attributes['g']['bool'].append(attr)
            attributes['g']['categorical'].append(attr)
        elif typ in ('i', 'f', 'u'):
            attributes['g']['numerical'].append(attr)
        elif typ == 'b':
            attributes['g']['bool'].append(attr)
            attributes['g']['categorical'].append(attr)

    attributes['c']['numerical'].extend([
        'n_genes',
        'n_counts',
    ])

    for attr in attributes['g']['bool']:
        attr2 = 'pct_counts_' + attr
        if attr2 not in adata.obs.columns:
            attr2 += '*'
        attributes['c']['numerical'].append(attr2)

    attributes['g']['numerical'].extend([
        'n_cells',
        'n_counts',
        'mean_counts',
        'pct_dropout_by_counts',
    ])
    logging.debug(attributes)
    return attributes


def _attributes_exists(name, attributes, dtype):
    cond_cat = ''
    if name.startswith('c:') or name.startswith('g:'):
        cond_cat, _, cond_name = name.partition(':')
        found = int(cond_name in attributes[cond_cat][dtype])
    else:
        cond_name = name
        if cond_name in attributes['c'][dtype]:
            cond_cat += 'c'
        if cond_name in attributes['g'][dtype]:
            cond_cat += 'g'
        found = len(cond_cat)
    return found, cond_cat, cond_name


def _get_filter_conditions(attributes, param, category, subset):
    conditions = {
        'c': {
            'numerical': [],
            'categorical': [],
            'bool': [],
        },
        'g': {
            'numerical': [],
            'categorical': [],
            'bool': [],
        },
    }
    percent_top_pattern = re.compile(r'^pct_counts_in_top_(?P<n>\d+)_genes$')
    pct_top = []
    qc_vars_pattern = re.compile(r'^pct_counts_(?P<qc_var>\S+)$')
    qc_vars = []

    for name, vmin, vmax in param:
        found, cond_cat, cond_name = _attributes_exists(
            name, attributes, 'numerical')
        pt_match = percent_top_pattern.match(cond_name)
        qv_match = qc_vars_pattern.match(cond_name)
        if found > 1:
            raise click.ClickException(f'Ambiguous parameter "{name}" found in '
                                       'both cell and gene table')
        if found < 1:
            if pt_match:
                pct_top.append(int(pt_match['n']))
                cond_cat = 'c'
            elif qv_match and qv_match['qc_var'] in attributes['g']['bool']:
                qc_vars.append(qv_match['qc_var'])
                cond_cat = 'c'
            else:
                raise click.ClickException(f'Parameter "{name}" unavailable')
        if pt_match or qv_match:
            vmin *= 100
            vmax *= 100
        conditions[cond_cat]['numerical'].append([cond_name, vmin, vmax])

    for name, values in category + subset:
        found, cond_cat, cond_name = _attributes_exists(
            name, attributes, 'categorical')
        if found > 1:
            raise click.ClickException(f'Ambiguous attribute "{name}" found in '
                                       'both cell and gene table')
        if found < 1:
            raise click.ClickException(f'Attribute "{name}" unavailable')
        if not isinstance(values, (list, tuple)):
            fh = values
            values = fh.read().rstrip().split('\n')
            fh.close()
        conditions[cond_cat]['categorical'].append((cond_name, values))

    logging.debug((conditions, qc_vars, pct_top))
    return conditions, qc_vars, sorted(pct_top)


def _repr_obj(obj, padding='  ', level=0):
    if isinstance(obj, dict):
        obj_str = '\n'.join(['\n'.join([
            padding * level + k + ':', _repr_obj(v, level=level+1)
        ]) for k, v in obj.items()])
    elif isinstance(obj, (tuple, list, set)):
        obj_str = '\n'.join([_repr_obj(elm, level=level) for elm in obj])
    else:
        obj_str = padding * level + repr(obj)
    return obj_str



"""
Provide helper functions for constructing sub-commands
"""

import click
import pandas as pd
import scanpy as sc
from .cmd_options import CMD_OPTIONS
from .lib._paga import plot_paga
from .obj_utils import _save_matrix

def make_subcmd(cmd_name, func, cmd_desc, arg_desc, opt_set = None):
    """
    Factory function that returns a sub-command function
    """
    opt_set = opt_set if opt_set else cmd_name
    options = CMD_OPTIONS[opt_set]
    option_spec = [click.command(cmd_name)]
    option_spec.extend(options)

    def add_docstring(cmd_desc, arg_desc):
        def docstring_dec(obj):
            obj.__doc__ = obj.__doc__.format(
                cmd_desc=cmd_desc, arg_desc=arg_desc)
            return obj
        return docstring_dec

    @add_options(option_spec)
    @add_docstring(cmd_desc, arg_desc)
    def cmd(
            input_obj=None,
            output_obj=None,
            input_format=None,
            output_format=None,
            zarr_chunk_size=None,
            loom_write_obsm_varm=False,
            export_mtx=None,
            show_obj=None,
            **kwargs
    ):
        """{cmd_desc}\n\n\b\n{arg_desc}"""
        if input_obj:
            adata = _read_obj(input_obj, input_format=input_format)
            func(adata, **kwargs)
        else:
            adata = func(**kwargs)

        if output_obj:
            _write_obj(
                adata,
                output_obj,
                output_format=output_format,
                chunk_size=zarr_chunk_size,
                write_obsm_varm=loom_write_obsm_varm,
                export_mtx=export_mtx,
                show_obj=show_obj,
            )
        return 0

    return cmd


def add_options(options):
    """
    Returns a decorator to group multiple click decorators
    """
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options


def _read_obj(input_obj, input_format='anndata', **kwargs):
    if input_format == 'anndata':
        adata = sc.read(input_obj, **kwargs)
    elif input_format == 'loom':
        adata = sc.read_loom(input_obj, **kwargs)
    else:
        raise NotImplementedError(
            'Unsupported input format: {}'.format(input_format))
    return adata


def _write_obj(
        adata,
        output_obj,
        output_format='anndata',
        chunk_size=None,
        export_mtx=None,
        show_obj=None,
        write_obsm_varm=False,
        **kwargs
):
    if output_format == 'anndata':
        adata.write(output_obj, compression='gzip')
    elif output_format == 'loom':
        adata.write_loom(output_obj, write_obsm_varm=write_obsm_varm )
    elif output_format == 'zarr':
        adata.write_zarr(output_obj, chunk_size=chunk_size, **kwargs)
    else:
        raise NotImplementedError(
            'Unsupported output format: {}'.format(output_format))
    if export_mtx:
        write_mtx(adata, fname_prefix=export_mtx, **kwargs)
    if show_obj:
        click.echo(adata, err=show_obj == 'stderr')
    return 0


def write_mtx(adata, fname_prefix='', var=None, obs=None, use_raw=False, use_layer=None):
    """Export AnnData object to mtx formt
    * Parameters
        + adata : AnnData
        An AnnData object
        + fname_prefix : str
        Prefix of the exported files. If not empty and not ending with '/' or '_',
        a '_' will be appended. Full names will be <fname_prefix>matrix.mtx,
        <fname_prefix>genes.tsv, <fname_prefix>barcodes.tsv
        + var : list
        A list of column names to be exported to gene table
        + obs : list
        A list of column names to be exported to barcode/cell table
    """
    if fname_prefix and not (fname_prefix.endswith('/') or fname_prefix.endswith('_')):
        fname_prefix = fname_prefix + '_'
    if var is None:
        var = []
    if obs is None:
        obs = []
    if use_raw:
        adata = adata.raw
    obs = list(set(obs) & set(adata.obs.columns))
    var = list(set(var) & set(adata.var.columns))

    import scipy.sparse as sp
    if use_layer is not None:
        mat=sp.coo_matrix(adata.layer[use_layer])
    else:
        mat = sp.coo_matrix(adata.X)
    
    n_obs, n_var = mat.shape
    n_entry = len(mat.data)
    header = '%%MatrixMarket matrix coordinate real general\n%\n{} {} {}\n'.format(
        n_var, n_obs, n_entry)
    df = pd.DataFrame({'col': mat.col + 1, 'row': mat.row + 1, 'data': mat.data})
    mtx_fname = fname_prefix + 'matrix.mtx'
    gene_fname = fname_prefix + 'genes.tsv'
    barcode_fname = fname_prefix + 'barcodes.tsv'
    with open(mtx_fname, 'a') as fh:
        fh.write(header)
        df.to_csv(fh, sep=' ', header=False, index=False)

    obs_df = adata.obs[obs].reset_index(level=0)
    obs_df.to_csv(barcode_fname, sep='\t', header=False, index=False)
    var_df = adata.var[var].reset_index(level=0)
    if not var:
        var_df['gene'] = var_df['index']
    var_df.to_csv(gene_fname, sep='\t', header=False, index=False)


def make_plot_function(func_name, kind=None):
    """Make plot function that handles common plotting parameters
    """

    # Provide a function translation

    plot_funcs = {
        'embedding': sc.pl.embedding,
        'scatter': sc.pl.scatter,
        'sviol': sc.pl.stacked_violin,
        'rgg_sviol': sc.pl.rank_genes_groups_stacked_violin,
        'dot': sc.pl.dotplot,
        'rgg_dot': sc.pl.rank_genes_groups_dotplot,
        'matrix': sc.pl.matrixplot,
        'rgg_matrix': sc.pl.rank_genes_groups_matrixplot,
        'heat': sc.pl.heatmap,
        'rgg_heat': sc.pl.rank_genes_groups_heatmap,
    }

    def plot_function(
            adata,
            output_fig=None,
            fig_size=None,
            fig_dpi=300,
            fig_fontsize=15,
            **kwargs,
    ):
        sc.settings.set_figure_params(dpi=fig_dpi, fontsize=fig_fontsize)
        if fig_size:
            from matplotlib import rcParams
            rcParams.update({'figure.figsize': fig_size})

        # Choose the function to run

        is_rgg=False

        if func_name in plot_funcs:
            if 'rgg' in kwargs:
                if kwargs['rgg'] == True:
                    is_rgg=True
                    func=plot_funcs[ 'rgg_' + func_name ]
                    kwargs.pop('var_names', None)
                else:
                    func = plot_funcs[ func_name ]
                    kwargs.pop('groups', None)
                    kwargs.pop('n_genes', None)
        
                kwargs.pop('rgg')
            else:
                func = plot_funcs[ func_name ]
        else:
            func = globals()[func_name]

        # Generate the output file name

        figname = False
        showfig = True
        if output_fig:
            import os
            import matplotlib.pyplot as plt
            sc.settings.figdir = os.path.dirname(output_fig) or '.'

            figname = os.path.basename(output_fig)
            showfig = False

        # Run the selected function

        func(
            adata,
            save=figname,
            show=showfig,
            **kwargs)

        # Rename output to the spefied file name. We need to work out what
        # prefix the function will have used for its output files.

        if output_fig:
            prefix=''
            if func_name == 'scatter' or func_name == 'embedding':
                prefix =  kwargs.get('basis', func.__name__)
            elif kind:
                prefix = kind 
            elif func_name in plot_funcs:
                prefix = plot_funcs[ func_name ].__name__.split('.')[-1]
                if func_name in [ 'sviol', 'rgg_sviol', 'dot', 'rgg_dot', 'matrix', 'rgg_matrix' ]:
                    prefix = prefix + '_'

            os.rename(
                os.path.join(sc.settings.figdir, prefix + figname), output_fig)
            plt.close()

    return plot_function

# Wrap matrix-processing functions in logic to back up .X or specified input
# layers prior to processing

def make_matrix_function(func): 

    def matrix_function(
        adata,
        save_raw=True, 
        save_layer=None, 
        **kwargs,
    ):

        # For the subset of matrix functions that allow layer specification,
        # pass that as the thing to save.
        
        layer=None
        if 'layer' in kwargs:
            layer = kwargs['layer']

        _save_matrix(adata, save_raw, save_layer=save_layer, layer=layer)
        func(
            adata,
            **kwargs
        )
        return adata

    return matrix_function



"""
scanpy fdg
"""

import scanpy as sc
from ..obj_utils import (
    _backup_obsm_key,
    _delete_obsm_backup_key,
    _rename_obsm_key,
    write_embedding,
)


def fdg(
        adata,
        layout='fa',
        key_added_ext=None,
        random_state=0,
        export_embedding=None,
        **kwargs
):
    """
    Wrapper function for sc.tl.draw_graph, for supporting named slot of fdg
    embeddings.
    """
    sc.tl.draw_graph(
        adata,
        layout=layout,
        key_added_ext=key_added_ext,
        random_state=random_state,
        **kwargs,
    )

    fdg_key  = f'X_draw_graph_{key_added_ext or layout}'

    if export_embedding is not None:
        write_embedding(adata, fdg_key, export_embedding, key_added=key_added_ext)
    return adata



"""
Provides sub-commands
"""

import os
import sys
import scanpy as sc
import scanpy.external as sce

from .cmd_utils import (
    make_subcmd,
    make_plot_function,
    make_matrix_function,
)
from .lib._read import read_10x
from .lib._filter import filter_anndata
from .lib._norm import normalize
from .lib._hvg import hvg
from .lib._pca import pca
from .lib._neighbors import neighbors
from .lib._umap import umap
from .lib._tsne import tsne
from .lib._fdg import fdg
from .lib._louvain import louvain
from .lib._leiden import leiden
from .lib._diffexp import diffexp
from .lib._paga import paga
from .lib._diffmap import diffmap
from .lib._dpt import dpt
from .lib._bbknn import bbknn
from .lib._mnn import mnn_correct
from .lib._combat import combat

LANG = os.environ.get('LANG', None)

if LANG is None or not (LANG.endswith('UTF-8') or
                        LANG.endswith('UTF8') or
                        LANG.endswith('utf-8') or
                        LANG.endswith('utf8')):
    print('This programme requires a UTF-8 locale, please check your $LANG setting.')
    sys.exit(0)


_I_DESC = '<input_obj>:   input file in format specfied by --input-format'
_O_DESC = '<output_obj>:  output file in format specfied by --output-format'
_P_DESC = '<output_fig>:  output figure in pdf or png format'
_IO_DESC = '\n'.join([_I_DESC, _O_DESC])
_IP_DESC = '\n'.join([_I_DESC, _P_DESC])


READ_CMD = make_subcmd(
    'read',
    read_10x,
    cmd_desc='Read 10x data and save in specified format.',
    arg_desc=_O_DESC,
)


FILTER_CMD = make_subcmd(
    'filter',
    make_matrix_function(filter_anndata),
    cmd_desc='Filter data based on specified conditions.',
    arg_desc=_IO_DESC,
)


NORM_CMD = make_subcmd(
    'norm',
    make_matrix_function(normalize),
    cmd_desc='Normalise data per cell.',
    arg_desc=_IO_DESC,
)


HVG_CMD = make_subcmd(
    'hvg',
    hvg,
    cmd_desc='Find highly variable genes.',
    arg_desc=_IO_DESC,
)


SCALE_CMD = make_subcmd(
    'scale',
    make_matrix_function(sc.pp.scale),
    cmd_desc='Scale data per gene.',
    arg_desc=_IO_DESC,
)


REGRESS_CMD = make_subcmd(
    'regress',
    make_matrix_function(sc.pp.regress_out),
    cmd_desc='Regress-out observation variables.',
    arg_desc=_IO_DESC,
)


PCA_CMD = make_subcmd(
    'pca',
    pca,
    cmd_desc='Dimensionality reduction by PCA.',
    arg_desc=_IO_DESC,
)

NEIGHBOR_CMD = make_subcmd(
    'neighbor',
    neighbors,
    cmd_desc='Compute a neighbourhood graph of observations.',
    arg_desc=_IO_DESC,
)

UMAP_CMD = make_subcmd(
    'umap',
    umap,
    cmd_desc='Embed the neighborhood graph using UMAP.',
    arg_desc=_IO_DESC,
)

TSNE_CMD = make_subcmd(
    'tsne',
    tsne,
    cmd_desc='Embed the cells using t-SNE.',
    arg_desc=_IO_DESC,
)

FDG_CMD = make_subcmd(
    'fdg',
    fdg,
    cmd_desc='Embed the neighborhood graph using force-directed graph.',
    arg_desc=_IO_DESC,
)

DIFFMAP_CMD = make_subcmd(
    'diffmap',
    diffmap,
    cmd_desc='Embed the neighborhood graph using diffusion map.',
    arg_desc=_IO_DESC,
)

LOUVAIN_CMD = make_subcmd(
    'louvain',
    louvain,
    cmd_desc='Find clusters by Louvain algorithm.',
    arg_desc=_IO_DESC,
)

LEIDEN_CMD = make_subcmd(
    'leiden',
    leiden,
    cmd_desc='Find clusters by Leiden algorithm.',
    arg_desc=_IO_DESC,
)

DIFFEXP_CMD = make_subcmd(
    'diffexp',
    diffexp,
    cmd_desc='Find markers for each clusters.',
    arg_desc=_IO_DESC,
)

PAGA_CMD = make_subcmd(
    'paga',
    paga,
    cmd_desc='Trajectory inference by abstract graph analysis.',
    arg_desc=_IO_DESC,
)

DPT_CMD = make_subcmd(
    'dpt',
    dpt,
    cmd_desc='Calculate diffusion pseudotime relative to the root cells.',
    arg_desc=_IO_DESC,
)

PLOT_EMBED_CMD = make_subcmd(
    'embed',
    make_plot_function('embedding'),
    cmd_desc='Plot cell embeddings.',
    arg_desc=_IP_DESC,
)

PLOT_STACKED_VIOLIN_CMD = make_subcmd(
    'sviol',
    make_plot_function('sviol'),
    cmd_desc='Plot stacked violin plots.',
    arg_desc=_IP_DESC,
)

PLOT_DOT_CMD = make_subcmd(
    'dot',
    make_plot_function('dot'),
    cmd_desc='Plot a dot plot of expression values.',
    arg_desc=_IP_DESC,
)

PLOT_MATRIX_CMD = make_subcmd(
    'matrix',
    make_plot_function('matrix'),
    cmd_desc='Plot a heatmap of the mean expression values per cluster.',
    arg_desc=_IP_DESC,
)

PLOT_HEATMAP_CMD = make_subcmd(
    'heat',
    make_plot_function('heat'),
    cmd_desc='Plot a heatmap of the expression values of genes.',
    arg_desc=_IP_DESC,
)

PLOT_PAGA_CMD = make_subcmd(
    'paga',
    make_plot_function('plot_paga', kind='paga'),
    cmd_desc='Plot PAGA trajectories.',
    arg_desc=_IP_DESC,
    opt_set='plot_paga'
)

COMBAT_CMD = make_subcmd(
    'combat',
    combat,
    cmd_desc='ComBat function for batch effect correction',
    arg_desc=_IO_DESC
)

HARMONY_INTEGRATE_CMD = make_subcmd(
    'harmony',
    sce.pp.harmony_integrate,
    cmd_desc='Use harmonypy [Korunsky19] to integrate different experiments.',    
    arg_desc=_IO_DESC,
)

BBKNN_CMD = make_subcmd(
    'bbknn',
    bbknn,
    cmd_desc='Batch balanced kNN [Polanski19].',
    arg_desc=_IO_DESC,
)

MNN_CORRECT_CMD = make_subcmd(
    'mnn',
    make_matrix_function(mnn_correct),
    cmd_desc='Correct batch effects by matching mutual nearest neighbors [Haghverdi18] [Kang18].',
    arg_desc=_IO_DESC,
)



"""
scanpy hvg
"""

import numpy as np
import scanpy as sc

def hvg(
        adata,
        mean_limits=(0.0125, 3),
        disp_limits=(0.5, float('inf')),
        **kwargs,
):
    """
    Wrapper function for sc.highly_variable_genes()
    """

    # Check for n_top_genes beeing greater than the total genes

    if 'n_top_genes' in kwargs and kwargs['n_top_genes'] is not None:
        kwargs['n_top_genes'] = min(adata.n_vars, kwargs['n_top_genes'])
    
    sc.pp.highly_variable_genes(
        adata,
        min_mean=mean_limits[0],
        max_mean=mean_limits[1],
        min_disp=disp_limits[0],
        max_disp=disp_limits[1],
        **kwargs,
    )
    
    return adata



"""
Provide helper functions for command line parsing with click
"""

import click


class NaturalOrderGroup(click.Group):
    """Command group trying to list subcommands in the order they were added.

    With decorator, use::

        @click.group(cls=NaturalOrderGroup)
    """
    def list_commands(self, ctx):
        """List command names as they are in commands dict.

        If the dict is OrderedDict, it will preserve the order commands
        were added.
        """
        return self.commands.keys()


class CommaSeparatedText(click.ParamType):
    """
    Comma separated text
    """
    def __init__(self, dtype=click.STRING, simplify=False, length=None):
        self.dtype = dtype
        self.dtype_name = _get_type_name(dtype)
        self.simplify = simplify
        self.length = length
        if length and length <= 3:
            self.name = ','.join([f'{self.dtype_name}'] * length)
        else:
            self.name = '{}[,{}...]'.format(self.dtype_name, self.dtype_name)

    def convert(self, value, param, ctx):
        try:
            if value is None:
                converted = None
            else:
                converted = list(map(self.dtype, str(value).split(',')))
                if self.simplify and len(converted) == 1:
                    converted = converted[0]
        except ValueError:
            self.fail(
                '{} is not a valid comma separated list of {}'.format(
                    value, self.dtype_name),
                param,
                ctx
            )
        if self.length:
            if len(converted) != self.length:
                self.fail(
                    '{} is not a valid comma separated list of length {}'.format(
                        value, self.length),
                    param,
                    ctx
                )
        return converted


class Dictionary(click.ParamType):
    """
    Text to be parsed as a python dict definition
    """
    def __init__(self, keys=None):
        self.name = 'TEXT:VAL[,TEXT:VAL...]'
        self.keys = keys

    def convert(self, value, param, ctx):
        try:
            converted = dict()
            for token in value.split(','):
                if ':' not in token:
                    raise ValueError
                key, _, value = token.partition(':')
                if not key:
                    raise ValueError
                if isinstance(self.keys, (list, tuple)) and key not in self.keys:
                    self.fail(f'{key} is not a valid key ({self.keys})')
                if value == 'None':
                    value = None
                elif value.lower() == 'true':
                    value = True
                elif value.lower() == 'false':
                    value = False
                else:
                    try:
                        value = float(value)
                    except ValueError:
                        pass
                converted[key] = value
            return converted
        except ValueError:
            self.fail(
                f'{value} is not a valid python dict definition',
                param,
                ctx
            )


def _get_type_name(obj):
    name = 'text'
    try:
        name = getattr(obj, 'name')
    except AttributeError:
        name = getattr(obj, '__name__')
    return name


def valid_limit(ctx, param, value):
    if value[0] > value[1]:
        param.type.fail(
            'lower limit must not exceed upper limit', param, ctx)
    return value


def valid_parameter_limits(ctx, param, value):
    for val in value:
        if val[1] > val[2]:
            param.type.fail(
                'lower limit must not exceed upper limit', param, ctx)
    return value


def mutually_exclusive_with(param_name):
    internal_name = param_name.strip('-').replace('-', '_').lower()
    def valid_mutually_exclusive(ctx, param, value):
        try:
            other_value = ctx.params[internal_name]
        except KeyError:
            return value
        if (value is None) == (other_value is None):
            param.type.fail(
                'mutually exclusive with "{}", one and only one must be '
                'specified.'.format(param_name),
                param,
                ctx,
            )
        return value
    return valid_mutually_exclusive


def required_by(param_name):
    internal_name = param_name.strip('-').replace('-', '_').lower()
    def required(ctx, param, value):
        try:
            other_value = ctx.params[internal_name]
        except KeyError:
            return value
        if other_value and not value:
            param.type.fail('required by "{}".'.format(param_name), param, ctx,)
        return value
    return required



"""
Provides exported functions
"""

from ._read import read_10x
from ._filter import filter_anndata
from ._norm import normalize
from ._hvg import hvg
from ._neighbors import neighbors
from ._umap import umap
from ._fdg import fdg
from ._tsne import tsne
from ._louvain import louvain
from ._leiden import leiden
from ._diffexp import diffexp, diffexp_paired, extract_de_table
from ._diffmap import diffmap
from ._dpt import dpt
from ._paga import paga, plot_paga
from ..cmd_utils import _read_obj as read_obj
from ..cmd_utils import _write_obj as write_obj



"""
scanpy diffexp
"""

import pandas as pd
import scanpy as sc
import logging

def diffexp(
        adata,
        use_raw=True,
        n_genes=None,
        key_added='rank_genes_groups',
        layer=None,
        logreg_param=None,
        filter_params=None,
        save=None,
        groupby=None,
        groups=None,
        **kwargs,
):
    """
    Wrapper function for sc.tl.rank_genes_groups.
    """
    if adata.raw is None:
        use_raw = False

    if n_genes is None:
        n_genes = adata.raw.shape[1] if use_raw else adata.shape[1]

    if logreg_param and isinstance(logreg_param, dict):
        for key, val in logreg_param:
            kwargs[key] = val

    key_added = key_added if key_added else 'rank_genes_groups'
    diff_key = (key_added + f'_{layer}') if layer else key_added

    if groups == 'all':

        # Avoid divisions by zeros for singlet groups. See 
        # https://github.com/theislab/scanpy/pull/1490#issuecomment-726031442.
        
        groups_to_test = list(
            adata.obs[groupby]
            .value_counts()
            .loc[lambda x: x > 1]
            .index
        )

        if len(groups_to_test) < len(adata.obs[groupby].cat.categories):
            groups = groups_to_test
            logging.warning('Singlet groups removed before passing to rank_genes_groups()')

    sc.tl.rank_genes_groups(
        adata, use_raw=use_raw, n_genes=n_genes, key_added=diff_key, groupby=groupby, groups = groups, **kwargs)

    de_tbl = extract_de_table(adata.uns[diff_key])

    if isinstance(filter_params, dict):
        sc.tl.filter_rank_genes_groups(
            adata,
            key=diff_key,
            key_added=diff_key + '_filtered',
            use_raw=use_raw,
            **filter_params,
        )

        de_tbl = extract_de_table(adata.uns[diff_key + '_filtered'])
        de_tbl = de_tbl.loc[de_tbl.genes.astype(str) != 'nan', :]

    if save:
        de_tbl.to_csv(save, sep='\t', header=True, index=False)

    return de_tbl


def diffexp_paired(adata, groupby, pair, **kwargs):
    """
    Restrict DE to between a pair of clusters, return both up and down genes
    """
    test, ref = pair
    de_key = f'de.{test}-{ref}'
    up_de = diffexp(
        adata,
        key_added=de_key,
        groupby=groupby,
        groups=[test],
        reference=ref,
        **kwargs,
    )
    ref, test = pair
    de_key = f'de.{test}-{ref}'
    down_de = diffexp(
        adata,
        key_added=de_key,
        groupby=groupby,
        groups=[test],
        reference=ref,
        **kwargs,
    )
    return up_de, down_de


def extract_de_table(de_dict):
    """
    Extract DE table from adata.uns
    """
    if de_dict['params']['method'] == 'logreg':
        requested_fields = ('scores',)
    else:
        requested_fields = ('scores', 'logfoldchanges', 'pvals', 'pvals_adj',)
    gene_df = _recarray_to_dataframe(de_dict['names'], 'genes')[
        ['cluster', 'rank', 'genes']]
    gene_df['ref'] = de_dict['params']['reference']
    gene_df = gene_df[['cluster', 'ref', 'rank', 'genes']]
    de_df = pd.DataFrame({
        field: _recarray_to_dataframe(de_dict[field], field)[field]
        for field in requested_fields if field in de_dict
    })
    return gene_df.merge(de_df, left_index=True, right_index=True)


def _recarray_to_dataframe(array, field_name):
    return pd.DataFrame(array).reset_index().rename(
        columns={'index': 'rank'}).melt(
            id_vars='rank', var_name='cluster', value_name=field_name)



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='scanpy-scripts',
    version='0.4.0',
    author='nh3',
    author_email='nh3@users.noreply.github.com',
    description='Scripts for using scanpy from the command line',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ebi-gene-expression-group/scanpy-scripts',
    packages=find_packages(),
    scripts=[
        'scanpy-scripts-tests.bats',
    ],
    entry_points=dict(
        console_scripts=[
            'scanpy-cli=scanpy_scripts.cli:cli',
            'scanpy-read-10x=scanpy_scripts.cmds:READ_CMD',
            'scanpy-filter-cells=scanpy_scripts.cmds:FILTER_CMD',
            'scanpy-filter-genes=scanpy_scripts.cmds:FILTER_CMD',
            'scanpy-normalise-data=scanpy_scripts.cmds:NORM_CMD',
            'scanpy-find-variable-genes=scanpy_scripts.cmds:HVG_CMD',
            'scanpy-scale-data=scanpy_scripts.cmds:SCALE_CMD',
            'scanpy-regress=scanpy_scripts.cmds:REGRESS_CMD',
            'scanpy-run-pca=scanpy_scripts.cmds:PCA_CMD',
            'scanpy-neighbors=scanpy_scripts.cmds:NEIGHBOR_CMD',
            'scanpy-run-tsne=scanpy_scripts.cmds:TSNE_CMD',
            'scanpy-run-umap=scanpy_scripts.cmds:UMAP_CMD',
            'scanpy-find-cluster=scanpy_scripts.cli:cluster',
            'scanpy-find-markers=scanpy_scripts.cmds:DIFFEXP_CMD',
        ]
    ),
    install_requires=[
        'packaging',
        'anndata',
        'scipy',
        'matplotlib',
        'pandas',
        'h5py<3.0.0',
        'scanpy>=1.6.0',
        'louvain',
        'leidenalg',
        'loompy>=2.0.0,<3.0.0',
        'MulticoreTSNE',
        'Click',
        'umap-learn<0.4.0',
        'harmonypy>=0.0.5',
        'bbknn>=1.3.12',
        'mnnpy>=0.1.9.5'
    ],
)



"""
Provides read_10x()
"""

import pandas as pd
import scanpy as sc


def read_10x(
        input_10x_h5,
        input_10x_mtx,
        genome='hg19',
        var_names='gene_symbols',
        extra_obs=None,
        extra_var=None
):
    """
    Wrapper function for sc.read_10x_h5() and sc.read_10x_mtx(), mainly to
    support adding extra metadata
    """
    if input_10x_h5 is not None:
        adata = sc.read_10x_h5(input_10x_h5, genome=genome)
    elif input_10x_mtx is not None:
        adata = sc.read_10x_mtx(input_10x_mtx, var_names=var_names)

    if extra_obs:
        obs_tbl = pd.read_csv(extra_obs, sep='\t', header=0, index_col=0)
        adata.obs = adata.obs.merge(
            obs_tbl,
            how='left',
            left_index=True,
            right_index=True,
            suffixes=(False, False),
        )

    if extra_var:
        var_tbl = pd.read_csv(extra_var, sep='\t', header=0, index_col=0)
        adata.var = adata.var.merge(
            var_tbl,
            how='left',
            left_index=True,
            right_index=True,
            suffixes=(False, False),
        )
    return adata



"""
scanpy umap
"""

import scanpy as sc
from ..obj_utils import (
    _set_default_key,
    _restore_default_key,
    _backup_obsm_key,
    _rename_obsm_key,
    _delete_obsm_backup_key,
    write_embedding
)

def umap(
        adata,
        key_added=None,
        random_state=0,
        export_embedding=None,
        **kwargs,
):
    """
    Wrapper function for sc.tl.umap, for supporting named slot of umap embeddings
    """
    if not isinstance(random_state, (list, tuple)):
        _backup_obsm_key(adata, 'X_umap')

        sc.tl.umap(adata, random_state=random_state, **kwargs)

        umap_key = 'X_umap'
        if key_added:
            umap_key = f'X_umap_{key_added}'
            _rename_obsm_key(adata, 'X_umap', umap_key)
        else:
            _delete_obsm_backup_key(adata, 'X_umap')

        if export_embedding is not None:
            write_embedding(adata, umap_key, export_embedding, key_added=key_added)
    else:
        for i, rseed in enumerate(random_state):
            if key_added is None:
                umap_key = f'r{rseed}'
            elif not isinstance(key_added, (list, tuple)):
                umap_key = f'{key_added}_r{rseed}'
            elif len(key_added) == len(random_state):
                umap_key = key_added[i]
            else:
                raise ValueError('`key_added` can only be None, a scalar, or an '
                                 'iterable of the same length as `random_state`.')
            umap(
                adata,
                key_added=umap_key,
                random_state=rseed,
                **kwargs,
            )
    return adata



"""
scanpy external mnn
"""

import scanpy.external as sce
import numpy as np 
import click

# Wrapper for mnn allowing use of non-standard slot

def mnn_correct(adata, key=None, key_added=None, var_subset=None, layer=None, **kwargs):
    """
    Wrapper function for sce.pp.mnn_correct(), for supporting non-standard neighbors slot
    """

    # mnn will use .X, so we need to put other layers there for processing

    if layer:
        adata.layers['X_backup'] = adata.X
        adata.X = adata.layers[layer]

    # mnn_correct() wants batches in separate adatas

    batches = np.unique(adata.obs[key])
    alldata = []
    for batch in batches:
        alldata.append( adata[adata.obs[key] == batch,] )

    # Process var_subset into a list of strings that can be provided to
    # mnn_correct()

    if var_subset is not None and len(var_subset) > 0 and var_subset[0] is not None:
   
        subset = []
 
        for name, values in var_subset :
            if name in adata.var:
                if adata.var[name].dtype == 'bool':
                    values = [ True if x.lower() == "true" else x for x in values  ]
            else:
                raise click.ClickException(f'Var "{name}" unavailable')

            ind = [ x in values for x in adata.var[name]  ]
            subset = subset + adata.var.index[ ind  ].to_list()

        var_subset = set(subset)
        print('Will use %d selected genes for MNN' % len(var_subset))

    else:
        var_subset = None

    # Here's the main bit

    cdata = sce.pp.mnn_correct(*alldata, var_subset = var_subset, do_concatenate = True, index_unique = None, **kwargs)    
    
    # If user has specified key_added = X then they want us to overwrite .X,
    # othwerwise copy the .X to a named layer of the original object. In either
    # case make sure obs and var are the same as the original.

    if key_added is None or key_added != 'X':

        mnn_key = 'mnn'
        if layer:
            mnn_key = f"{mnn_key}_{layer}"
        
            # Layers is set (so we're not storing computed results in the .X,
            # and we had to overwrite .X to run mnn), and key_added shows we're
            # not storing in the .X, so we need to restore from the backup.

            adata.X = adata.layers['X_backup']

        if key_added:
            mnn_key = f"{mnn_key}_{key_added}"
   
        adata.layers[mnn_key] = cdata[0][adata.obs.index, adata.var.index].X

    else:
        adata.X = cdata[0][adata.obs.index, adata.var.index].X

    # Delete the backup of .X if we needed one 

    if layer:
        del adata.layers['X_backup'] 

    return adata



"""
scanpy diffmap
"""

import scanpy as sc
from ..obj_utils import (
    _rename_obsm_key,
    write_embedding,
)


def diffmap(
        adata,
        key_added=None,
        export_embedding=None,
        **kwargs,
):
    """
    Wrapper function for sc.tl.diffmap, for supporting named slot
    """
    sc.tl.diffmap(adata, **kwargs)
    
    diffmap_key = 'X_diffmap'
    if key_added:
        diffmap_key = f'{diffmap_key}_{key_added}'
        _rename_obsm_key(adata, 'X_diffmap', diffmap_key)

    if export_embedding is not None:
        write_embedding(adata, diffmap_key, export_embedding, key_added=key_added)
    return adata



"""
scanpy paga
"""

import numpy as np
import scanpy as sc
from ..obj_utils import (
    _backup_default_key,
    _delete_backup_key,
    _rename_default_key,
    _set_default_key,
    _restore_default_key,
)


def paga(
        adata,
        key_added=None,
        **kwargs,
):
    """
    Wrapper function for sc.tl.paga, for supporting named slot
    """
    sc.tl.paga(adata, **kwargs)

    if key_added:
        paga_key = f'paga_{key_added}'
        _rename_default_key(adata.uns, 'paga', paga_key)
    else:
        _delete_backup_key(adata.uns, 'paga')

    return adata


def plot_paga(
        adata,
        use_key='paga',
        basis=None,
        layout=None,
        init_pos=None,
        legend_loc='on data',
        color=None,
        size=None,
        title=None,
        show=None,
        save=None,
        **kwargs,
):
    """Make PAGA plot
    """
    if basis is not None and f'X_{basis}' in adata.obsm.keys():
        ax = sc.pl.embedding(
            adata,
            basis=basis,
            color=color,
            legend_loc=legend_loc,
            size=size,
            title=None,
            save=False,
            show=False,
        )

        grouping = adata.uns[use_key]['groups']
        categories = list(adata.obs[grouping].cat.categories)
        obsm = adata.obsm[f'X_{basis}']
        group_pos = np.zeros((len(categories), 2))
        for i, label in enumerate(categories):
            offset = 1 if basis.startswith('diffmap') else 0
            _scatter = obsm[adata.obs[grouping] == label, (0+offset):(2+offset)]
            x_pos, y_pos = np.median(_scatter, axis=0)
            group_pos[i] = [x_pos, y_pos]

        _set_default_key(adata.uns, 'paga', use_key)
        kwargs['node_size_scale'] = 0
        kwargs['fontsize'] = 1
        kwargs['pos'] = group_pos
        kwargs['color'] = None
        try:
            sc.pl.paga(
                adata,
                ax=ax,
                title=title,
                show=show,
                save=save,
                **kwargs,
            )
        finally:
            _restore_default_key(adata.uns, 'paga', use_key)
    else:
        _set_default_key(adata.uns, 'paga', use_key)
        try:
            sc.pl.paga(
                adata,
                layout=layout,
                init_pos=init_pos,
                title=title,
                show=show,
                save=save,
                **kwargs
            )
        finally:
            _restore_default_key(adata.uns, 'paga', use_key)

    return adata
